@@ -77,8 +77,8 @@
          :format-arguments format-arguments))
 
 (defmacro nconc-2 (a b)
-  (let ((tmp (sb-xc:gensym))
-        (tmp2 (sb-xc:gensym)))
+  (let ((tmp (gensym))
+        (tmp2 (gensym)))
     `(let ((,tmp ,a)
            (,tmp2 ,b))
        (if ,tmp
@@ -609,6 +609,23 @@
           (t (return (values form documentation declarations lambda-list))))
         finally (return (values nil documentation declarations lambda-list))))
 
+(defun make-interpreted-function
+      (&key name lambda-list env declarations documentation body source-location
+            (debug-lambda-list lambda-list))
+    (let ((function (%make-interpreted-function
+                     name name lambda-list debug-lambda-list env
+                     declarations documentation body source-location)))
+      (setf (%funcallable-instance-fun function)
+            #'(lambda (&rest args)
+                (interpreted-apply function args)))
+      function))
+
+(defmethod print-object ((obj interpreted-function) stream)
+  (print-unreadable-object (obj stream
+                            :identity (not (interpreted-function-name obj)))
+    (format stream "~A ~A" '#:interpreted-function
+            (interpreted-function-name obj))))
+
 ;;; Create an interpreted function from the lambda-form EXP evaluated
 ;;; in the environment ENV.
 (defun eval-lambda (exp env)
@@ -1012,6 +1029,8 @@
         (sb-sys:with-pinned-objects ((%eval (car values) env))
           (eval-with-pinned-objects (cons (cdr values) body) env)))))
 
+(defparameter *eval-level* -1)
+(defparameter *eval-verbose* nil)
 (defvar *eval-dispatch-functions* nil)
 
 ;;; Dispatch to the appropriate EVAL-FOO function based on the contents of EXP.
