@@ -165,7 +165,7 @@
 ;;; mutex is needed. More importantly the sigchld signal handler also
 ;;; accesses it, that's why we need without-interrupts.
 (defmacro with-active-processes-lock (() &body body)
-  `(sb-thread::with-system-mutex (*active-processes-lock*)
+  `(with-system-mutex (*active-processes-lock*)
      ,@body))
 
 (deftype process-status ()
@@ -307,7 +307,8 @@ PROCESS."
   "Hand SIGNAL to PROCESS. If WHOM is :PID, use the kill Unix system call. If
    WHOM is :PROCESS-GROUP, use the killpg Unix system call. If WHOM is
    :PTY-PROCESS-GROUP deliver the signal to whichever process group is
-   currently in the foreground."
+   currently in the foreground.
+   Returns T if successful, otherwise returns NIL and error number (two values)."
   (let ((pid (ecase whom
                ((:pid :process-group)
                 (process-pid process))
@@ -543,7 +544,10 @@ status slot."
         ;; Copy string.
         (copy-ub8-to-system-area octets 0 string-sap 0 size)
         ;; NULL-terminate it
-        (system-area-ub8-fill 0 string-sap size 4)
+        ;; (As it says up top: "assume 4 byte is enough for everyone.")
+        (let ((sap (sap+ string-sap size)))
+          (setf (sap-ref-8 sap 0) 0 (sap-ref-8 sap 1) 0
+                (sap-ref-8 sap 2) 0 (sap-ref-8 sap 3) 0))
         ;; Put the pointer in the vector.
         (setf (sap-ref-sap vec-sap vec-index-offset) string-sap)
         ;; Advance string-sap for the next string.
@@ -589,7 +593,8 @@ status slot."
   (envp (* c-string))
   (pty-name c-string)
   (channel (array int 2))
-  (dir c-string))
+  (dir c-string)
+  (preserve-fds (* int)))
 
 #-win32
 (define-alien-routine wait-for-exec
@@ -649,6 +654,11 @@ status slot."
       (loop for arg in args
             collect (coerce arg 'simple-string))))
 
+(defmacro coerce-or-copy (object type)
+  `(if (typep ,object ,type)
+       (copy-seq ,object)
+       (coerce ,object ,type)))
+
 ;;; FIXME: There shouldn't be two semiredundant versions of the
 ;;; documentation. Since this is a public extension function, the
 ;;; documentation should be in the doc string. So all information from
@@ -694,26 +704,34 @@ status slot."
 ;;; the fork worked, and NIL if it did not.
 (defun run-program (program args
                     &key
-                    (env nil env-p)
-                    (environment
-                     (when env-p
-                       (unix-environment-sbcl-from-cmucl env))
-                     environment-p)
-                    (wait t)
-                    search
-                    #-win32 pty
-                    input
-                    if-input-does-not-exist
-                    output
-                    (if-output-exists :error)
-                    (error :output)
-                    (if-error-exists :error)
-                    status-hook
-                    (external-format :default)
-                    directory
-                    #+win32 (escape-arguments t))
+                      (env nil env-p)
+                      (environment
+                       (when env-p
+                         (unix-environment-sbcl-from-cmucl env))
+                       environment-p)
+                      (wait t)
+                      search
+                      #-win32 pty
+                      input
+                      if-input-does-not-exist
+                      output
+                      (if-output-exists :error)
+                      (error :output)
+                      (if-error-exists :error)
+                      status-hook
+                      (external-format :default)
+                      directory
+                      preserve-fds
+                      #+win32 (escape-arguments t)
+                      #+win32 (window nil))
   "RUN-PROGRAM creates a new process specified by PROGRAM.
-ARGS are passed as the arguments to the program.
+ARGS is a list of strings to be passed literally to the new program.
+In POSIX environments, this list becomes the array supplied as the second
+parameter to the execv() or execvp() system call, each list element becoming
+one array element. The strings should not contain shell escaping, as there is
+no shell involvement. Further note that while conventionally the process
+receives its own pathname in argv[0], that is automatic, and the 0th string
+should not be present in ARGS.
 
 The program arguments and the environment are encoded using the
 default external format for streams.
@@ -799,9 +817,22 @@ Users Manual for details about the PROCESS structure.
       Specifies the directory in which the program should be run.
       NIL (the default) means the directory is unchanged.
 
+   :PRESERVE-FDS
+      A sequence of file descriptors which should remain open in the child
+      process.
+
    Windows specific options:
    :ESCAPE-ARGUMENTS (default T)
-      Controls escaping of the arguments passed to CreateProcess."
+      Controls escaping of the arguments passed to CreateProcess.
+   :WINDOW (default NIL)
+      When NIL, the subprocess decides how it will display its window. The
+      following options control how the subprocess window should be displayed:
+      :HIDE, :SHOW-NORMAL, :SHOW-MAXIMIZED, :SHOW-MINIMIZED, :SHOW-NO-ACTIVATE,
+      :SHOW-MIN-NO-ACTIVE, :SHOW-NA.
+      Note: console application subprocesses may or may not display a console
+      window depending on whether the SBCL runtime is itself a console or GUI
+      application. Invoke CMD /C START to consistently display a console window
+      or use the :WINDOW :HIDE option to consistently hide the console window."
   (when (and env-p environment-p)
     (error "can't specify :ENV and :ENVIRONMENT simultaneously"))
   (let* (;; Clear various specials used by GET-DESCRIPTOR-FOR to
@@ -855,11 +886,13 @@ Users Manual for details about the PROCESS structure.
                                       output cookie
                                       :direction :output
                                       :if-exists if-output-exists
+                                      :if-does-not-exist :create
                                       :external-format external-format)
                (with-fd-and-stream-for ((stderr error-stream)  :error
                                         error cookie
                                         :direction :output
                                         :if-exists if-error-exists
+                                        :if-does-not-exist :create
                                         :external-format external-format)
                  ;; Make sure we are not notified about the child
                  ;; death before we have installed the PROCESS
@@ -871,21 +904,32 @@ Users Manual for details about the PROCESS structure.
                        (with-open-pty ((pty-name pty-stream) (pty cookie))
                          (setf (values child #+win32 handle)
                                #+win32
-                               (sb-thread::with-system-mutex (*spawn-lock*)
+                               (with-system-mutex (*spawn-lock*)
                                  (sb-win32::mswin-spawn
                                   progname
                                   args
                                   stdin stdout stderr
-                                  search environment-vec directory))
+                                  search environment-vec directory
+                                  window
+                                  preserve-fds))
                                #-win32
-                               (with-args (args-vec args)
-                                 (sb-thread::with-system-mutex (*spawn-lock*)
-                                   (spawn progname args-vec
-                                          stdin stdout stderr
-                                          (if search 1 0)
-                                          environment-vec pty-name
-                                          channel
-                                          directory))))
+                               (let ((preserve-fds
+                                       (and preserve-fds
+                                            (sort (coerce-or-copy preserve-fds
+                                                                  '(simple-array (signed-byte #.(alien-size int)) (*)))
+                                                  #'<))))
+                                 (with-pinned-objects (preserve-fds)
+                                   (with-args (args-vec args)
+                                     (with-system-mutex (*spawn-lock*)
+                                       (spawn progname args-vec
+                                              stdin stdout stderr
+                                              (if search 1 0)
+                                              environment-vec pty-name
+                                              channel
+                                              directory
+                                              (if preserve-fds
+                                                  (vector-sap preserve-fds)
+                                                  (int-sap 0))))))))
                          (unless (minusp child)
                            #-win32
                            (setf child (wait-for-exec child channel))
@@ -1144,16 +1188,16 @@ Users Manual for details about the PROCESS structure.
                     (push write-fd *close-on-error*)
                     (let ((stream (make-fd-stream write-fd :output t
                                                          :element-type :default
-                                                         :external-format
-                                                         external-format)))
+                                                         :external-format external-format
+                                                         :auto-close t)))
                       (values read-fd stream)))
                  (:output
                     (push read-fd *close-on-error*)
                     (push write-fd *close-in-parent*)
                     (let ((stream (make-fd-stream read-fd :input t
                                                          :element-type :default
-                                                         :external-format
-                                                         external-format)))
+                                                         :external-format external-format
+                                                         :auto-close t)))
                       (values write-fd stream)))
                  (t
                     (sb-unix:unix-close read-fd)
@@ -1285,11 +1329,15 @@ Users Manual for details about the PROCESS structure.
   if not available."
   (or sb-sys::*software-version*
       (setf sb-sys::*software-version*
-            (string-trim '(#\newline)
-                         (%with-output-to-string (stream)
+            (possibly-base-stringize
+             #+linux
+             (with-open-file (f "/proc/sys/kernel/osrelease") (read-line f))
+             #-linux
+             (string-trim '(#\newline)
+                          (%with-output-to-string (stream)
                            (run-program "/bin/uname"
                                         ;; "-r" on haiku just prints "1"
                                         ;; but "-v" prints some detail.
                                         #+haiku '("-v")
                                         #-haiku '("-r")
-                                        :output stream))))))
+                                        :output stream)))))))
