@@ -61,7 +61,11 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
   (do-all-symbols (s)
     (let ((dd (sb-kernel:find-defstruct-description s nil)))
       (when (and dd (not (sb-kernel::dd-null-lexenv-p dd)))
-        (push (sb-kernel:dd-name dd) result))))
+        (let ((pkg (symbol-package (sb-kernel:dd-name dd))))
+          (unless (member (package-name pkg)
+                          '("SB-RBTREE.WORD" "SB-RBTREE.MAP")
+                          :test 'string=)
+            (push (sb-kernel:dd-name dd) result))))))
   (assert (equal result '(sb-c::conset))))
 
 ;;; Assert that genesis preserved shadowing symbols.
@@ -127,8 +131,8 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
 ;;;     ((:or :macro (:match "$EARLY-") (:match "$BOOT-"))
 ;;;     (declare (optimize (speed 0))))))
 ;;;
-(let ((sources (with-open-file (f (merge-pathnames "../../build-order.lisp-expr"
-                                                   *load-pathname*))
+(defvar *sbclroot* "")
+(let ((sources (with-open-file (f (merge-pathnames "build-order.lisp-expr" *load-pathname*))
                  (read f) ; skip over the make-host-{1,2} input files
                  (read f)))
       (sb-c::*handled-conditions* sb-c::*handled-conditions*))
@@ -143,7 +147,8 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                            ;; a literal (when compiling in the LOAD step)
                            t))
                 (output
-                  (compile-file-pathname stem
+                  (compile-file-pathname
+                   (concatenate 'string *sbclroot* stem)
                    :output-file
                    (merge-pathnames
                     (concatenate
@@ -170,7 +175,8 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                                     (safety 2) (speed 2)
                                     (sb-c:insert-step-conditions 0)
                                     (sb-c:alien-funcall-saves-fp-and-pc #+x86 3 #-x86 0)))
-                        (compile-file stem :output-file output)))
+                        (compile-file (concatenate 'string *sbclroot* stem)
+                                      :output-file output)))
                      ((nil) output))
                   (cond ((not output-truename)
                          (error "COMPILE-FILE of ~S failed." stem))
@@ -199,7 +205,9 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                               ((sb-kernel:redefinition-with-defgeneric
                                 #'muffle-warning))
                             (let ((sb-c::*source-namestring* fullname))
-                              (load output-truename)))
+                              ;; RISCV is slow, I'd like to see it doing something
+                              ;; rather than appearing to go out to lunch
+                              (load output-truename :verbose (or #+riscv t))))
                     (error "LOAD of ~S failed." output-truename))
                   (sb-int:/show "done loading" output-truename))))))))
 
@@ -219,15 +227,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                                            (write-to-string c :escape nil)))
                           (cerror "Finish warm compile ignoring the problem" c)))))
         (with-compilation-unit ()
-          (do-srcs group)
-          ;; I do not know why #+sb-show gets several "undefined-type CLASS" warnings
-          ;; that #-sb-show doesn't. And CLASS is a reserved name not defined as yet,
-          ;; so we have to pretend that didn't happen, otherwise the warning about
-          ;; not being able to define CLASS as a type name breaks the build.
-          #+sb-show
-          (setq sb-c::*undefined-warnings*
-                (delete 'class sb-c::*undefined-warnings*
-                        :key #'sb-c::undefined-warning-name))))))))
+          (do-srcs group)))))))
 
 (sb-c::dump/restore-interesting-types 'write)
 (when (hash-table-p sb-c::*static-vop-usage-counts*)
@@ -239,3 +239,12 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
         (push (cons (gethash name sb-c::*static-vop-usage-counts* 0) name) list))
       (dolist (cell (sort list #'> :key #'car))
         (format output "~7d ~s~%" (car cell) (cdr cell))))))
+
+(when (sb-sys:find-dynamic-foreign-symbol-address "tot_gc_nsec")
+  (let* ((run-sec (/ (get-internal-real-time) internal-time-units-per-second))
+         (gc-nsec (extern-alien "tot_gc_nsec" unsigned))
+         (gc-msec (/ (float gc-nsec) 1000000)))
+    (format t "~&Done with warm.lisp. INTERNAL-REAL-TIME=~Fs~@[, GC=~Fms (~,1,2f%)~]~%"
+            run-sec
+            (if (plusp gc-msec) gc-msec) ; timing wasn't enabled if this is 0
+            (/ gc-msec (* 1000 run-sec)))))
