@@ -14,51 +14,31 @@
 (defconstant arg-count-sc (make-sc+offset immediate-arg-scn nargs-offset))
 (defconstant closure-sc (make-sc+offset descriptor-reg-sc-number lexenv-offset))
 
-;;; Make a passing location TN for a local call return PC.  If
-;;; standard is true, then use the standard (full call) location,
-;;; otherwise use any legal location.  Even in the non-standard case,
-;;; this may be restricted by a desire to use a subroutine call
-;;; instruction.
-(defun make-return-pc-passing-location (standard)
-  (declare (ignore standard))
-  (make-wired-tn *backend-t-primitive-type* control-stack-sc-number
-                 lra-save-offset))
-
 (defconstant return-pc-passing-offset
   (make-sc+offset control-stack-sc-number lra-save-offset))
 
-;;; This is similar to MAKE-RETURN-PC-PASSING-LOCATION, but makes a
-;;; location to pass OLD-FP in.
-;;;
-;;; This is wired in both the standard and the local-call conventions,
-;;; because we want to be able to assume it's always there. Besides,
-;;; the ARM doesn't have enough registers to really make it profitable
-;;; to pass it in a register.
-(defun make-old-fp-passing-location ()
-  (make-wired-tn *fixnum-primitive-type* control-stack-sc-number
-                 ocfp-save-offset))
-
 (defconstant old-fp-passing-offset
   (make-sc+offset control-stack-sc-number ocfp-save-offset))
 
 ;;; Make the TNs used to hold OLD-FP and RETURN-PC within the current
 ;;; function. We treat these specially so that the debugger can find
 ;;; them at a known location.
-(defun make-old-fp-save-location (env)
+(defun make-old-fp-save-location ()
   ;; Unlike the other backends, ARM function calling is designed to
   ;; pass OLD-FP within the stack frame rather than in a register.  As
   ;; such, in order for lifetime analysis not to screw up, we need it
   ;; to be a stack TN wired to the save offset, not a normal TN with a
   ;; wired SAVE-TN.
-  (physenv-debug-live-tn (make-wired-tn *fixnum-primitive-type*
-                                        control-stack-arg-scn
-                                        ocfp-save-offset)
-                         env))
-(defun make-return-pc-save-location (physenv)
-  (physenv-debug-live-tn
-   (make-wired-tn *backend-t-primitive-type* control-stack-sc-number
-                  lra-save-offset)
-   physenv))
+  (let ((tn (make-wired-tn *fixnum-primitive-type*
+                           control-stack-arg-scn
+                           ocfp-save-offset)))
+    (setf (tn-kind tn) :environment)
+    tn))
+(defun make-return-pc-save-location ()
+  (let ((tn (make-wired-tn *backend-t-primitive-type* control-stack-sc-number
+                           lra-save-offset)))
+    (setf (tn-kind tn) :environment)
+    tn))
 
 ;;; Make a TN for the standard argument count passing location.  We
 ;;; only need to make the standard location, since a count is never
@@ -112,8 +92,8 @@
     (store-stack-offset value frame-pointer variable-home-tn)))
 
 (define-vop (xep-allocate-frame)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:info start-lab)
-  (:temporary (:scs (interior-reg)) lip)
   (:generator 1
     ;; Make sure the function is aligned, and drop a label pointing to this
     ;; function header.
@@ -122,7 +102,7 @@
     ;; Allocate function header.
     (inst simple-fun-header-word)
     (inst .skip (* (1- simple-fun-insts-offset) n-word-bytes))
-    (inst str lip (@ cfp-tn (* lra-save-offset n-word-bytes)))))
+    (inst str lr (@ cfp-tn (* lra-save-offset n-word-bytes)))))
 
 (define-vop (xep-setup-sp)
   (:vop-var vop)
@@ -141,9 +121,13 @@
   (:info callee)
   (:generator 2
     (move res csp-tn)
-    (inst add csp-tn csp-tn (add-sub-immediate
-                             (* (max 1 (sb-allocated-size 'control-stack)) n-word-bytes)))
-    (when (ir2-physenv-number-stack-p callee)
+    (let ((size (add-sub-immediate (* (max 1 (sb-allocated-size 'control-stack)) n-word-bytes))))
+      (cond ((typep size '(signed-byte 9))
+             (inst str cfp-tn (@ csp-tn size :post-index)))
+            (t
+             (inst add csp-tn csp-tn size)
+             (storew cfp-tn res ocfp-save-offset))))
+    (when (ir2-environment-number-stack-p callee)
       (inst sub nfp nsp-tn (add-sub-immediate
                             (bytes-needed-for-non-descriptor-stack-frame)))
       (inst mov-sp nsp-tn nfp))))
@@ -151,17 +135,21 @@
 ;;; Allocate a partial frame for passing stack arguments in a full call.  Nargs
 ;;; is the number of arguments passed.  If no stack arguments are passed, then
 ;;; we don't have to do anything.
+;;; LR and CFP are always saved on the stack, but it's safe to have two words above CSP.
 (define-vop (allocate-full-call-frame)
   (:info nargs)
   (:results (res :scs (any-reg)))
   (:generator 2
-    (move res csp-tn)
-        (let ((size (add-sub-immediate (* (max 2 nargs) n-word-bytes))))
-      (cond ((typep size '(signed-byte 9))
-             (inst str cfp-tn (@ csp-tn size :post-index)))
-            (t
-             (inst add csp-tn csp-tn size)
-             (storew cfp-tn res ocfp-save-offset))))))
+    (if (<= nargs register-arg-count)
+        ;; Don't touch RES, the call vops would use CSP-TN in this case.
+        (storew cfp-tn csp-tn ocfp-save-offset)
+        (let ((size (add-sub-immediate (* nargs n-word-bytes))))
+          (move res csp-tn)
+          (cond ((typep size '(signed-byte 9))
+                 (inst str cfp-tn (@ csp-tn size :post-index)))
+                (t
+                 (inst add csp-tn csp-tn size)
+                 (storew cfp-tn res ocfp-save-offset)))))))
 
 ;;; Emit code needed at the return-point from an unknown-values call
 ;;; for a fixed number of values.  VALUES is the head of the TN-REF
@@ -189,10 +177,12 @@
 ;;;  -- Reset SP.  This must be done whenever other than 1 value is returned,
 ;;;     regardless of the number of values desired.
 
-(defun default-unknown-values (vop values nvals move-temp)
+(defun default-unknown-values (vop values nvals move-temp node)
   (declare (type (or tn-ref null) values)
            (type unsigned-byte nvals) (type tn move-temp))
-  (let ((expecting-values-on-stack (> nvals register-arg-count)))
+  (let* ((type (sb-c::node-derived-type node))
+         (min-values (values-type-min-value-count type))
+         (expecting-values-on-stack (> nvals register-arg-count)))
     (note-this-location vop (if (<= nvals 1)
                                 :single-value-return
                                 :unknown-return))
@@ -207,13 +197,22 @@
              (val (tn-ref-across values) (tn-ref-across val)))
             ((= i (min nvals register-arg-count)))
           (unless (eq (tn-kind (tn-ref-tn val)) :unused)
-           (inst csel (tn-ref-tn val) null-tn (tn-ref-tn val) :ne))))
+            (cond
+              ((and min-values
+                    (> min-values i)))
+              (t
+               (inst csel (tn-ref-tn val) null-tn (tn-ref-tn val) :ne))))))
 
       ;; If we're not expecting values on the stack, all that
       ;; remains is to clear the stack frame (for the multiple-
       ;; value return case).
-      (unless expecting-values-on-stack
-        (inst csel csp-tn ocfp-tn csp-tn :eq))
+      (unless (or expecting-values-on-stack
+                  (type-single-value-p type))
+        (cond ((values-type-may-be-single-value-p type)
+               (inst csel csp-tn ocfp-tn csp-tn :eq))
+              ((eq type *empty-type*))
+              (t
+               (inst mov csp-tn ocfp-tn))))
 
       ;; If we ARE expecting values on the stack, we need to
       ;; either move them to their result location or to set their
@@ -223,9 +222,10 @@
         ;; For the single-value return case, fake up NARGS and
         ;; OCFP so that we don't screw ourselves with the
         ;; defaulting and stack clearing logic.
-        (inst csel ocfp-tn csp-tn ocfp-tn :ne)
-        (inst mov tmp-tn (fixnumize 1))
-        (inst csel nargs-tn tmp-tn nargs-tn :ne)
+        (unless (> min-values 1)
+          (inst csel ocfp-tn csp-tn ocfp-tn :ne)
+          (inst mov tmp-tn (fixnumize 1))
+          (inst csel nargs-tn tmp-tn nargs-tn :ne))
 
         ;; For each expected stack value...
         (do ((i register-arg-count (1+ i))
@@ -236,22 +236,64 @@
                   (tn-ref-across val)))
             ((null val))
           (let ((tn (tn-ref-tn val)))
-            (if (eq (tn-kind tn) :unused)
-                (incf decrement (fixnumize 1))
-                (assemble ()
-                  ;; ... Load it if there is a stack value available, or
-                  ;; default it if there isn't.
-                  (inst subs nargs-tn nargs-tn decrement)
-                  (setf decrement (fixnumize 1))
-                  (inst b :lt NONE)
-                  (loadw move-temp ocfp-tn i 0)
-                  NONE
-                  (sc-case tn
-                    (control-stack
-                     (inst csel move-temp null-tn move-temp :lt)
-                     (store-stack-tn tn move-temp))
-                    (t
-                     (inst csel tn null-tn move-temp :lt)))))))
+            (cond ((eq (tn-kind tn) :unused)
+                   (incf decrement (fixnumize 1)))
+                  ((< i min-values)
+                   (incf decrement (fixnumize 1))
+                   (sc-case tn
+                     (control-stack
+                      (let* ((next (and (< (1+ i) min-values)
+                                        (tn-ref-across val)))
+                             (next-tn (and next
+                                           (tn-ref-tn next))))
+                        (cond ((and next-tn
+                                    (not (sc-is next-tn control-stack))
+                                    (neq (tn-kind next-tn) :unused)
+                                    (ldp-stp-offset-p (* i n-word-bytes) n-word-bits))
+                               (inst ldp move-temp next-tn
+                                     (@ ocfp-tn (* i n-word-bytes)))
+                               (store-stack-tn tn move-temp)
+                               (setf val next)
+                               (incf i)
+                               (incf decrement (fixnumize 1)))
+                              (t
+                               (loadw move-temp ocfp-tn i)
+                               (store-stack-tn tn move-temp)))))
+                     (t
+                      (let* ((next (and (< (1+ i) min-values)
+                                        (tn-ref-across val)))
+                             (next-tn (and next
+                                           (tn-ref-tn next))))
+                        (cond ((and next-tn
+                                    (neq (tn-kind next-tn) :unused)
+                                    (ldp-stp-offset-p (* i n-word-bytes) n-word-bits))
+                               (let ((stack (sc-is next-tn control-stack)))
+                                 (inst ldp tn (if stack
+                                                  move-temp
+                                                  next-tn)
+                                       (@ ocfp-tn (* i n-word-bytes)))
+                                 (when stack
+                                   (store-stack-tn next-tn move-temp)))
+                               (setf val next)
+                               (incf i)
+                               (incf decrement (fixnumize 1)))
+                              (t
+                               (loadw tn ocfp-tn i)))))))
+                  (t
+                   (assemble ()
+                     ;; ... Load it if there is a stack value available, or
+                     ;; default it if there isn't.
+                     (inst subs nargs-tn nargs-tn decrement)
+                     (setf decrement (fixnumize 1))
+                     (inst b :lt NONE)
+                     (loadw move-temp ocfp-tn i)
+                     NONE
+                     (sc-case tn
+                       (control-stack
+                        (inst csel move-temp null-tn move-temp :lt)
+                        (store-stack-tn tn move-temp))
+                       (t
+                        (inst csel tn null-tn move-temp :lt))))))))
         ;; Deallocate the callee stack frame.
         (move csp-tn ocfp-tn))))
   (values))
@@ -274,25 +316,38 @@
 ;;;    Args and Nargs are TNs wired to the named locations.  We must
 ;;; explicitly allocate these TNs, since their lifetimes overlap with the
 ;;; results Start and Count (also, it's nice to be able to target them).
-(defun receive-unknown-values (args nargs start count)
+(defun receive-unknown-values (node args nargs start count)
   (declare (type tn args nargs start count))
-  (assemble ()
-    (inst b :eq MULTIPLE)
-    (move start csp-tn)
-    (inst add csp-tn csp-tn n-word-bytes)
-    (inst str (first *register-arg-tns*) (@ start))
-    (inst mov count (fixnumize 1))
-    (inst b DONE)
-    MULTIPLE
-    #.(assert (evenp register-arg-count))
-    (do ((arg *register-arg-tns* (cddr arg))
-         (i 0 (+ i 2)))
-        ((null arg))
-      (inst stp (first arg) (second arg)
-            (@ args (* i n-word-bytes))))
-    (move start args)
-    (move count nargs)
-    DONE))
+  (let ((unused-count-p (eq (tn-kind count) :unused))
+        (unused-start-p (eq (tn-kind start) :unused))
+        (type (sb-c::node-derived-type node)))
+    (if (type-single-value-p type)
+        (assemble ()
+          (unless unused-start-p
+            (move start csp-tn))
+          (unless unused-count-p
+            (inst mov count (fixnumize 1)))
+          (inst str (first *register-arg-tns*) (@ csp-tn n-word-bytes :post-index)))
+        (assemble ()
+          (inst b :eq MULTIPLE)
+          (unless unused-start-p
+            (move start csp-tn))
+          (inst str (first *register-arg-tns*) (@ csp-tn n-word-bytes :post-index))
+          (unless unused-count-p
+            (inst mov count (fixnumize 1)))
+          (inst b DONE)
+          MULTIPLE
+          #.(assert (evenp register-arg-count))
+          (do ((arg *register-arg-tns* (cddr arg))
+               (i 0 (+ i 2)))
+              ((null arg))
+            (inst stp (first arg) (second arg)
+                  (@ args (* i n-word-bytes))))
+          (unless unused-start-p
+            (move start args))
+          (unless unused-count-p
+            (move count nargs))
+          DONE))))
 
 ;;; VOP that can be inherited by unknown values receivers.  The main
 ;;; thing this handles is allocation of the result temporaries.
@@ -514,12 +569,11 @@
   (:args (context-arg :target context :scs (descriptor-reg))
          (count-arg :target count :scs (any-reg)))
   (:arg-types * tagged-num)
-  (:temporary (:scs (any-reg) :from (:argument 0)) context)
+  (:temporary (:scs (descriptor-reg) :from (:argument 0)) context)
   (:temporary (:scs (any-reg) :from (:argument 1)) count)
   (:temporary (:scs (descriptor-reg) :from :eval) temp)
   (:temporary (:scs (any-reg) :from :eval) dst)
-  (:temporary (:sc non-descriptor-reg) pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:results (result :scs (descriptor-reg)))
   (:policy :safe)
   (:node-var node)
@@ -530,8 +584,9 @@
     (move result null-tn)
     (inst cbz count DONE)
 
-    (let ((dx-p (node-stack-allocate-p node)))
-      (pseudo-atomic (pa-flag :sync nil :elide-if dx-p)
+    (let ((dx-p (node-stack-allocate-p node))
+          (leave-pa (gen-label)))
+      (pseudo-atomic (lr :sync nil :elide-if dx-p)
         ;; Allocate a cons (2 words) for each item.
         (let ((size (cond (dx-p
                            (lsl count (1+ (- word-shift n-fixnum-tag-bits))))
@@ -539,9 +594,16 @@
                            (inst lsl temp count (1+ (- word-shift n-fixnum-tag-bits)))
                            temp))))
           (allocation 'list size list-pointer-lowtag dst
-                      :flag-tn pa-flag
+                      :flag-tn lr
                       :stack-allocate-p dx-p
-                      :lip lip))
+                      :overflow
+                      (lambda ()
+                        ;; The size will be computed by subtracting from CSP
+                        (inst mov tmp-tn context)
+                        (load-inline-constant lr `(:fixup listify-&rest :assembly-routine))
+                        (inst blr lr)
+                        (inst mov result tmp-tn)
+                        (inst b leave-pa))))
         (move result dst)
 
         (inst b ENTER)
@@ -562,7 +624,8 @@
         (inst b :gt LOOP)
 
         ;; NIL out the last cons.
-        (storew null-tn dst 1 list-pointer-lowtag)))
+        (storew null-tn dst 1 list-pointer-lowtag)
+        (emit-label LEAVE-PA)))
     DONE))
 
 ;;; Return the location and size of the more arg glob created by
@@ -647,6 +710,7 @@
   (:move-args :local-call)
   (:info arg-locs callee target nvals)
   (:vop-var vop)
+  (:node-var node)
   (:temporary (:scs (descriptor-reg) :from (:eval 0)) move-temp)
   (:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)
   (:temporary (:sc any-reg :offset ocfp-offset :from (:eval 0)) ocfp)
@@ -661,7 +725,7 @@
       (maybe-load-stack-tn cfp-tn fp)
       (note-this-location vop :call-site)
       (inst bl target)
-      (default-unknown-values vop values nvals move-temp)
+      (default-unknown-values vop values nvals move-temp node)
       (when cur-nfp
         (load-stack-tn cur-nfp nfp-save)))))
 
@@ -682,6 +746,7 @@
   (:info save callee target)
   (:ignore args save r0-temp)
   (:vop-var vop)
+  (:node-var node)
   (:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)
   (:generator 20
     (let ((cur-nfp (current-nfp-tn vop)))
@@ -695,7 +760,7 @@
       (note-this-location vop :call-site)
       (inst bl target)
       (note-this-location vop :unknown-return)
-      (receive-unknown-values values-start nvals start count)
+      (receive-unknown-values node values-start nvals start count)
       (when cur-nfp
         (load-stack-tn cur-nfp nfp-save)))))
 
@@ -740,25 +805,22 @@
 ;;; registers may be tied up by the more operand.  Instead, we use
 ;;; MAYBE-LOAD-STACK-TN.
 (define-vop (known-return)
-  (:args (old-fp :target old-fp-temp)
+  (:args (old-fp)
          (return-pc)
          (vals :more t))
-  (:temporary (:sc any-reg :from (:argument 0)) old-fp-temp)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:move-args :known-return)
   (:info val-locs)
-  (:ignore val-locs vals)
+  (:ignore old-fp return-pc val-locs vals)
   (:vop-var vop)
   (:generator 6
-              (maybe-load-stack-tn old-fp-temp old-fp)
-              (maybe-load-stack-tn lip return-pc)
-              (move csp-tn cfp-tn)
-              (let ((cur-nfp (current-nfp-tn vop)))
-                (when cur-nfp
-                  (inst add nsp-tn cur-nfp (add-sub-immediate
-                                            (bytes-needed-for-non-descriptor-stack-frame)))))
-              (move cfp-tn old-fp-temp)
-              (lisp-return lip :known)))
+    (move csp-tn cfp-tn)
+    (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
+    (let ((cur-nfp (current-nfp-tn vop)))
+      (when cur-nfp
+        (inst add nsp-tn cur-nfp (add-sub-immediate
+                                  (bytes-needed-for-non-descriptor-stack-frame)))))
+    (lisp-return lr :known)))
 
 ;;;; Full call:
 ;;;
@@ -799,7 +861,6 @@
 ;;; In tail call with fixed arguments, the passing locations are passed as a
 ;;; more arg, but there is no new-FP, since the arguments have been set up in
 ;;; the current frame.
-;(defvar fun-type :function)
 (defmacro define-full-call (name named return variable)
   (aver (not (and variable (eq return :tail))))
   `(define-vop (,name
@@ -831,6 +892,7 @@
          '((:move-args :full-call)))
 
      (:vop-var vop)
+     (:node-var node)
      (:info ,@(unless (or variable (eq return :tail)) '(arg-locs))
             ,@(unless variable '(nargs))
             ,@(when (eq named :direct) '(fun))
@@ -844,7 +906,7 @@
       ,@(unless variable '(args))
       ,@(ecase return
           (:fixed '(ocfp-temp))
-          (:tail '(old-fp return-pc))
+          (:tail '(old-fp return-pc node))
           (:unknown '(r0-temp))))
 
      ,@(unless (eq named :direct)
@@ -853,10 +915,6 @@
                         :to :eval)
                        ,(if named 'name-pass 'lexenv))))
 
-     ,@(unless named
-         `((:temporary (:scs (descriptor-reg) :to :eval)
-                       function)))
-
      (:temporary (:sc any-reg :offset nargs-offset :to
                       ,(if (eq return :fixed)
                            :save
@@ -877,7 +935,7 @@
      ,@(unless (eq return :tail)
          '((:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)))
 
-     (:temporary (:scs (interior-reg)) lip)
+     (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
 
      (:generator ,(+ (if named 5 0)
                      (if variable 19 1)
@@ -901,15 +959,7 @@
                               (ecase what
                                 (:load-nargs
                                  ,@(if variable
-                                       `((move nargs-pass csp-tn)
-                                         ;; The variable args are on the stack
-                                         ;; and become the frame, but there may
-                                         ;; be <4 args and 2 stack slots are
-                                         ;; assumed allocate on the call. So
-                                         ;; need to ensure there are at least 2
-                                         ;; slots. This just adds 2 more.
-                                         (inst add csp-tn nargs-pass (* 2 n-word-bytes))
-                                         (inst sub nargs-pass nargs-pass new-fp)
+                                       `((inst sub nargs-pass csp-tn new-fp)
                                          (inst asr nargs-pass nargs-pass (- word-shift n-fixnum-tag-bits))
                                          ,@(do ((arg *register-arg-names* (cddr arg))
                                                 (i 0 (+ i 2))
@@ -920,7 +970,8 @@
                                                           (@ new-fp ,(* i n-word-bytes)))
                                                    insts))
                                          (storew cfp-tn new-fp ocfp-save-offset))
-                                       '((load-immediate-word nargs-pass (fixnumize nargs)))))
+                                       '((unless (consp nargs)
+                                           (load-immediate-word nargs-pass (fixnumize nargs))))))
                                 ,@(if (eq return :tail)
                                       '((:frob-nfp
                                          (inst add nsp-tn cur-nfp (add-sub-immediate
@@ -928,7 +979,14 @@
                                       `((:frob-nfp
                                          (store-stack-tn nfp-save cur-nfp))
                                         (:load-fp
-                                         (move cfp-tn new-fp))))
+                                         (move cfp-tn (cond ,@(and
+                                                               (not variable)
+                                                               '(((<= (if (consp nargs)
+                                                                          (car nargs)
+                                                                          nargs) register-arg-count)
+                                                                  csp-tn)))
+                                                            (t
+                                                             new-fp))))))
                                 ((nil)))))
                           (insert-step-instrumenting ()
                             ;; Conditionally insert a conditional trap:
@@ -975,32 +1033,32 @@
                           (return)))
                      ,@(case named
                          ((t)
-                          `((loadw lip name-pass fdefn-raw-addr-slot other-pointer-lowtag)
+                          `((loadw lr name-pass fdefn-raw-addr-slot other-pointer-lowtag)
                             ,(if (eq return :tail)
-                                 `(inst add lip lip 4))))
+                                 `(inst add lr lr 4))))
                          (:direct
-                          `((inst ldr lip (@ null-tn (load-store-offset (static-fun-offset fun))))
+                          `((inst ldr lr (@ null-tn (load-store-offset (static-fun-offset fun))))
                             ,(if (eq return :tail)
-                                 `(inst add lip lip 4)))))
+                                 `(inst add lr lr 4)))))
 
                      (note-this-location vop :call-site)
 
                      ,(if named
                           (if (eq return :tail)
-                              `(inst br lip)
-                              `(inst blr lip))
+                              `(inst br lr)
+                              `(inst blr lr))
                           (if (eq return :tail)
-                              `(tail-call-unnamed lexenv function lip fun-type)
-                              `(call-unnamed lexenv function lip fun-type))))
+                              `(tail-call-unnamed lexenv lr fun-type)
+                              `(call-unnamed lexenv lr fun-type))))
 
                    ,@(ecase return
                        (:fixed
-                        '((default-unknown-values vop values nvals move-temp)
+                        '((default-unknown-values vop values nvals move-temp node)
                           (when cur-nfp
                             (load-stack-tn cur-nfp nfp-save))))
                        (:unknown
                         '((note-this-location vop :unknown-return)
-                          (receive-unknown-values values-start nvals start count)
+                          (receive-unknown-values node values-start nvals start count)
                           (when cur-nfp
                             (load-stack-tn cur-nfp nfp-save))))
                        (:tail))))))
@@ -1029,7 +1087,6 @@
   (:info fun-type)
   (:temporary (:sc any-reg :offset nl2-offset :from (:argument 0)) args)
   (:temporary (:sc descriptor-reg :offset lexenv-offset :from (:argument 1)) lexenv)
-  (:temporary (:scs (interior-reg)) lip)
   (:ignore old-fp-arg lra-arg)
   (:vop-var vop)
   (:generator 75
@@ -1044,12 +1101,11 @@
     (load-inline-constant tmp-tn
                           (if (eq fun-type :function)
                               '(:fixup tail-call-variable :assembly-routine)
-                              '(:fixup tail-call-callable-variable :assembly-routine))
-                          lip)
+                              '(:fixup tail-call-callable-variable :assembly-routine)))
     (inst br tmp-tn)))
 
 ;;; Invoke the function-designator FUN.
-(defun tail-call-unnamed (lexenv fun lip type)
+(defun tail-call-unnamed (lexenv lr type)
   (case type
     (:symbol
      (load-inline-constant tmp-tn '(:fixup tail-call-symbol :assembly-routine))
@@ -1063,14 +1119,11 @@
          (load-inline-constant tmp-tn '(:fixup tail-call-symbol :assembly-routine))
          (inst br tmp-tn))
        call
-       (loadw fun lexenv closure-fun-slot fun-pointer-lowtag)
-       (inst add lip fun
-             (+ (- (ash simple-fun-insts-offset word-shift)
-                   fun-pointer-lowtag)
-                4))
-       (inst br lip)))))
+       (loadw lr lexenv closure-fun-slot fun-pointer-lowtag)
+       (inst add lr lr 4)
+       (inst br lr)))))
 
-(defun call-unnamed (lexenv fun lip type)
+(defun call-unnamed (lexenv lr type)
   (case type
     (:symbol
      (load-inline-constant tmp-tn '(:fixup call-symbol :assembly-routine))
@@ -1085,11 +1138,8 @@
          (inst blr tmp-tn)
          (inst b ret))
        call
-       (loadw fun lexenv closure-fun-slot fun-pointer-lowtag)
-       (inst add lip fun
-             (- (ash simple-fun-insts-offset word-shift)
-                fun-pointer-lowtag))
-       (inst blr lip)
+       (loadw lr lexenv closure-fun-slot fun-pointer-lowtag)
+       (inst blr lr)
        ret))))
 
 ;;;; Unknown values return:
@@ -1099,7 +1149,7 @@
   (:args (old-fp)
          (return-pc)
          (value))
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:ignore value old-fp return-pc)
   (:vop-var vop)
   (:generator 6
@@ -1111,11 +1161,11 @@
     ;; Interrupts leave two words of space for the new frame, so it's safe
     ;; to deallocate the frame before accessing OCFP/LR.
     (move csp-tn cfp-tn)
-    (loadw-pair cfp-tn ocfp-save-offset lip lra-save-offset cfp-tn)
+    (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
     ;; Clear the control stack, and restore the frame pointer.
 
     ;; Out of here.
-    (lisp-return lip :single-value)))
+    (lisp-return lr :single-value)))
 
 ;;; Do unknown-values return of a fixed number of values.  The Values are
 ;;; required to be set up in the standard passing locations.  Nvals is the
@@ -1140,7 +1190,7 @@
   (:temporary (:sc descriptor-reg :offset r1-offset :from (:eval 0)) r1)
   (:temporary (:sc descriptor-reg :offset r2-offset :from (:eval 0)) r2)
   (:temporary (:sc descriptor-reg :offset r3-offset :from (:eval 0)) r3)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:temporary (:sc any-reg :offset nargs-offset) nargs)
   (:temporary (:sc any-reg :offset ocfp-offset) val-ptr)
   (:vop-var vop)
@@ -1153,15 +1203,15 @@
     (cond ((= nvals 1)
            ;; Clear the control stack, and restore the frame pointer.
            (move csp-tn cfp-tn)
-           (loadw-pair cfp-tn ocfp-save-offset lip lra-save-offset cfp-tn)
+           (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
            ;; Out of here.
-           (lisp-return lip :single-value))
+           (lisp-return lr :single-value))
           (t
            ;; Establish the values pointer.
            (move val-ptr cfp-tn)
            ;; restore the frame pointer and clear as much of the control
            ;; stack as possible.
-           (loadw-pair cfp-tn ocfp-save-offset lip lra-save-offset cfp-tn)
+           (loadw-pair cfp-tn ocfp-save-offset lr lra-save-offset cfp-tn)
            (inst add csp-tn val-ptr (add-sub-immediate (* nvals n-word-bytes)))
            ;; Establish the values count.
            (load-immediate-word nargs (fixnumize nvals))
@@ -1170,7 +1220,7 @@
              (dolist (reg (subseq (list r0 r1 r2 r3) nvals))
                (move reg null-tn)))
            ;; And away we go.
-           (lisp-return lip :multiple-values)))))
+           (lisp-return lr :multiple-values)))))
 
 ;;; Do unknown-values return of an arbitrary number of values (passed
 ;;; on the stack.)  We check for the common case of a single return
@@ -1187,10 +1237,10 @@
   (:temporary (:sc any-reg :offset nl1-offset :from (:argument 2)) vals)
   (:temporary (:sc any-reg :offset nargs-offset :from (:argument 3)) nvals)
   (:temporary (:sc descriptor-reg :offset r0-offset) r0)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg :offset lr-offset) lr)
   (:vop-var vop)
   (:generator 13
-    (maybe-load-stack-tn lip lra-arg)
+    (maybe-load-stack-tn lr lra-arg)
     ;; Clear the number stack.
     (let ((cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
@@ -1205,13 +1255,13 @@
     (inst ldr r0 (@ vals-arg))
     (move csp-tn cfp-tn)
     (move cfp-tn old-fp-arg)
-    (lisp-return lip :single-value)
+    (lisp-return lr :single-value)
 
     NOT-SINGLE
     (move old-fp old-fp-arg)
     (move vals vals-arg)
     (move nvals nvals-arg)
-    (load-inline-constant tmp-tn '(:fixup return-multiple :assembly-routine) lip)
+    (load-inline-constant tmp-tn '(:fixup return-multiple :assembly-routine))
     (inst br tmp-tn)))
 
 ;;; Single-stepping
