@@ -453,7 +453,10 @@
 ;;; source level because our fixed-values conventions are optimized
 ;;; for the common MV-BIND case.
 (defun make-type-check-form (types cast)
-  (let ((temps (make-gensym-list (length types))))
+  (let* ((temps (make-gensym-list (length types)))
+         (context (cast-context cast))
+         (restart (and (eq context :restart)
+                       (setf context (make-restart-location)))))
     (lambda (dummy)
       `(multiple-value-bind ,temps ,dummy
          ,@(mapcar
@@ -464,15 +467,18 @@
                          (let ((*unparse-fun-type-simplify* t))
                            (type-specifier type-to-check)))
                        (test (if not `(not ,spec) spec)))
-                  `(unless
-                       ,(with-ir1-environment-from-node cast ;; it performs its own inlining of SATISFIES
-                          (%source-transform-typep temp test))
-                     ,(internal-type-error-call temp
-                                                (if (fun-designator-type-p type-to-report)
-                                                    ;; Simplify
-                                                    (specifier-type 'function-designator)
-                                                    type-to-report)
-                                                (cast-context cast))))))
+                  `(progn
+                     (unless
+                         ,(with-ir1-environment-from-node cast ;; it performs its own inlining of SATISFIES
+                            (%source-transform-typep temp test))
+                       ,(internal-type-error-call temp
+                                                  (if (fun-designator-type-p type-to-report)
+                                                      ;; Simplify
+                                                      (specifier-type 'function-designator)
+                                                      type-to-report)
+                                                  context))
+                     ,@(and restart
+                            `((restart-point ,restart)))))))
             temps
             types)
          (values ,@temps)))))
@@ -495,39 +501,35 @@
   (let* ((lvar (node-lvar cast))
          (dest (and lvar (lvar-dest lvar)))
          (value (cast-value cast))
-         (atype (cast-asserted-type cast))
-         (condition 'type-warning)
-         (not-ok-uses '()))
+         (atype (cast-asserted-type cast)))
     (do-uses (use value)
       (let ((dtype (node-derived-type use)))
-        (if (values-types-equal-or-intersect dtype atype)
-            (setf condition 'type-style-warning)
-            (push use not-ok-uses))))
-    (dolist (use (nreverse not-ok-uses))
-      (let* ((*compiler-error-context* use)
-             (dtype (node-derived-type use))
-             (what (when (and (combination-p dest)
-                              (eq (combination-kind dest) :local))
-                     (let ((lambda (combination-lambda dest))
-                           (pos (position-or-lose
-                                 lvar (combination-args dest))))
-                       (format nil "~:[A possible~;The~] binding of ~S"
-                               (and (lvar-has-single-use-p lvar)
-                                    (eq (functional-kind lambda) :let))
-                               (leaf-source-name (elt (lambda-vars lambda)
-                                                      pos)))))))
-        (cond ((and (ref-p use) (constant-p (ref-leaf use)))
-               (warn condition
-                     :format-control "~:[This~;~:*~A~] is not a ~
+        (unless (or (values-types-equal-or-intersect dtype atype)
+                    (cast-mismatch-from-inlined-p cast use))
+          (let* ((*compiler-error-context* use)
+                 (dtype (node-derived-type use))
+                 (what (when (and (combination-p dest)
+                                  (eq (combination-kind dest) :local))
+                         (let ((lambda (combination-lambda dest))
+                               (pos (position-or-lose
+                                     lvar (combination-args dest))))
+                           (format nil "~:[A possible~;The~] binding of ~S"
+                                   (and (lvar-has-single-use-p lvar)
+                                        (eq (functional-kind lambda) :let))
+                                   (leaf-source-name (elt (lambda-vars lambda)
+                                                          pos)))))))
+            (cond ((and (ref-p use) (constant-p (ref-leaf use)))
+                   (warn 'type-style-warning
+                         :format-control "~:[This~;~:*~A~] is not a ~
                        ~<~%~9T~:;~/sb-impl:print-type/:~>~% ~S"
-                     :format-arguments
-                     (list what atype (constant-value (ref-leaf use)))))
-              (t
-               (warn condition
-                     :format-control
-                      "~:[Result~;~:*~A~] is a ~/sb-impl:print-type/, ~
+                         :format-arguments
+                         (list what atype (constant-value (ref-leaf use)))))
+                  (t
+                   (warn 'type-style-warning
+                         :format-control
+                         "~:[Result~;~:*~A~] is a ~/sb-impl:print-type/, ~
                        ~<~%~9T~:;not a ~/sb-impl:print-type/.~>"
-                     :format-arguments (list what dtype atype)))))))
+                         :format-arguments (list what dtype atype)))))))))
   (values))
 
 ;;; Loop over all blocks in COMPONENT that have TYPE-CHECK set,
