@@ -43,7 +43,6 @@
   ;; of SYMBOL-GLOBAL-VALUE.
   (similar-table (make-similarity-table) :type hash-table :read-only t)
   (eq-table (make-hash-table :test 'eq) :type hash-table :read-only t)
-  (named-constant-table (make-hash-table :test #'eq) :type hash-table :read-only t)
   ;; the INSTANCE table maps dumpable instances to unique IDs for calculating
   ;; a similarity hash of composite objects that contain instances.
   ;; A user-defined hash function can not use address-based hashing, and it is
@@ -89,6 +88,8 @@
   ;; a hash table of structures that are allowed to be dumped. If we
   ;; try to dump a structure that isn't in this hash table, we lose.
   (valid-structures (make-hash-table :test 'eq) :type hash-table)
+  ;; a hash table of slots to be saved when dumping an instance.
+  (saved-slot-names (make-hash-table :test 'eq) :type hash-table)
   ;; DEBUG-SOURCE written at the very beginning
   (source-info nil :type (or null sb-c::debug-source)))
 (declaim (freeze-type fasl-output))
@@ -212,11 +213,13 @@
 ;;; This structure holds information about a circularity.
 (defstruct (circularity (:copier nil))
   ;; the kind of modification to make to create circularity
-  (type (missing-arg) :type (member :rplaca :rplacd :svset :struct-set))
+  (type (missing-arg) :type (member :rplaca :rplacd :svset :struct-set :slot-set))
   ;; the object containing circularity
   object
   ;; index in object for circularity
   (index (missing-arg) :type index)
+  ;; slot name in object for circularity
+  (slot-name nil :type symbol)
   ;; the object to be stored at INDEX in OBJECT. This is that the key
   ;; that we were using when we discovered the circularity.
   value
@@ -469,7 +472,11 @@
                               ((eq x sb-c::*debug-name-ellipsis*) 2)
                               (t (bug "Bogus debug name marker")))))
              (instance
-              (dump-structure x file)
+              (multiple-value-bind (slot-names slot-names-p)
+                  (gethash x (fasl-output-saved-slot-names file))
+                (if slot-names-p
+                    (dump-instance-saving-slots x slot-names file)
+                    (dump-structure x file)))
               (eq-save-object x file))
              (array
               ;; DUMP-ARRAY (and its callees) are responsible for
@@ -506,7 +513,7 @@
 #+(and (not sb-xc-host) sb-simd-pack-256)
 (defun dump-simd-pack-256 (x file)
   (dump-fop 'fop-simd-pack file)
-  (dump-integer-as-n-bytes (logior (%simd-pack-256-tag x) 4) 8 file)
+  (dump-integer-as-n-bytes (logior (%simd-pack-256-tag x) (ash 1 6)) 8 file)
   (dump-integer-as-n-bytes (%simd-pack-256-0 x) 8 file)
   (dump-integer-as-n-bytes (%simd-pack-256-1 x) 8 file)
   (dump-integer-as-n-bytes (%simd-pack-256-2 x) 8 file)
@@ -564,7 +571,10 @@
                    (:rplaca     #.(get 'fop-rplaca 'opcode))
                    (:rplacd     #.(get 'fop-rplacd 'opcode))
                    (:svset      #.(get 'fop-svset 'opcode))
-                   (:struct-set #.(get 'fop-structset 'opcode)))
+                   (:struct-set #.(get 'fop-structset 'opcode))
+                   (:slot-set
+                    (dump-object (circularity-slot-name info) file)
+                    #.(get 'fop-slotset 'opcode)))
                  file)
       (dump-varint (gethash (circularity-object info) table) file)
       (dump-varint (circularity-index info) file))))
@@ -623,26 +633,17 @@
     (setf (gethash constant table) handle))
   (values))
 
-;;; Return T iff the constant named by NAME has already been
-;;; dumped. It's been dumped if it's in the NAMED-CONSTANTS table.
-(defun fasl-named-constant-already-dumped-p (name file)
-  (and (gethash name (fasl-output-named-constant-table file)) t))
-
-;;; Use HANDLE whenever we try to dump a constant named NAME. HANDLE
-;;; should have been returned earlier by
-;;; FASL-DUMP-LOAD-TIME-VALUE-LAMBDA.
-(defun fasl-note-handle-for-named-constant (name handle file)
-  (let ((table (fasl-output-named-constant-table file)))
-    (when (gethash name table)
-      (error "named constant ~S already dumped?" name))
-    (setf (gethash name table) handle))
-  (values))
-
 ;;; Note that the specified structure can just be dumped by
 ;;; enumerating the slots.
-(defun fasl-note-dumpable-instance (structure file)
+(defun fasl-validate-structure (structure file)
   (setf (gethash structure (fasl-output-valid-structures file)) t)
   (values))
+
+;;; Note that the specified standard object can just be dumped by
+;;; saving its slot values.
+(defun fasl-note-instance-saves-slots (instance slot-names file)
+  (setf (gethash instance (fasl-output-saved-slot-names file)) slot-names)
+  (values))
 
 ;;;; number dumping
 
@@ -893,8 +894,9 @@
        (dump-simple-vector simple-version file)
        (eq-save-object x file))
       (t
-       (dump-specialized-vector simple-version file)
-       (eq-save-object x file)))))
+       (unless (similar-check-table x file)
+         (dump-specialized-vector simple-version file)
+         (similar-save-object x file))))))
 
 ;;; Dump a SIMPLE-VECTOR, handling any circularities.
 (defun dump-simple-vector (v file)
@@ -1099,11 +1101,6 @@
       (dump-integer info fasl-output))
     (incf n)))
 
-;;; Dump a reference to the constant named by NAME.
-(defun dump-named-constant-reference (name fasl-output)
-  (dump-push (gethash name (fasl-output-named-constant-table fasl-output))
-             fasl-output))
-
 ;;; Dump out the constant pool and code-vector for component, push the
 ;;; result in the table, and return the offset.
 ;;;
@@ -1126,22 +1123,19 @@
          (header-length (length constants))
          (n-named-calls 0))
     (dump-object alloc-points fasl-output)
-    (collect ((patches))
+    (collect ((patches)
+              (named-constants))
       ;; Dump the constants, noting any :ENTRY constants that have to
       ;; be patched.
       (loop for i from sb-vm:code-constants-offset below header-length do
         (let ((entry (aref constants i)))
           (etypecase entry
             (constant
-             (if (and (sb-c::leaf-has-source-name-p entry)
-                      ;; We can't really reference constants defined
-                      ;; by name at load time in the same block
-                      ;; compilation unit, so dump it anonymously when
-                      ;; such a situation arises.
-                      (not (member (sb-c::leaf-source-name entry)
-                                   sb-c::*hairy-defconstants*)))
-                 (dump-named-constant-reference (sb-c::leaf-source-name entry) fasl-output)
-                 (dump-object (sb-c::constant-value entry) fasl-output)))
+             (cond ((sb-c::leaf-has-source-name-p entry)
+                    (named-constants (cons (sb-c::leaf-source-name entry) i))
+                    (dump-fop 'fop-misc-trap fasl-output))
+                   (t
+                    (dump-object (sb-c::constant-value entry) fasl-output))))
             (cons
              (ecase (car entry)
                (:constant ; anything that has not been wrapped in a #<CONSTANT>
@@ -1166,7 +1160,13 @@
                 (dump-fop 'fop-fdefn fasl-output))
                (:known-fun
                 (dump-object (cadr entry) fasl-output)
-                (dump-fop 'fop-known-fun fasl-output))))
+                (dump-fop 'fop-known-fun fasl-output))
+               (:coverage-marks
+                ;; Avoid the coalescence done by DUMP-VECTOR
+                (dump-specialized-vector (make-array (cdr entry)
+                                                     :element-type '(unsigned-byte 8)
+                                                     :initial-element #xFF)
+                                         fasl-output))))
             (null
              (dump-fop 'fop-misc-trap fasl-output)))))
 
@@ -1195,6 +1195,10 @@
           (push (cons handle (cdr patch))
                 (gethash (car patch)
                          (fasl-output-patch-table fasl-output))))
+        (dolist (named-constant (named-constants))
+          (dump-object (car named-constant) fasl-output)
+          (dump-push handle fasl-output)
+          (dump-fop 'fop-named-constant-set fasl-output (cdr named-constant)))
         handle))))
 
 ;;; This is only called from assemfile, which doesn't exist in the target.
@@ -1239,7 +1243,7 @@
   #+sb-dyncount
   (let ((info (sb-c::ir2-component-dyncount-info (component-info component))))
     (when info
-      (fasl-note-dumpable-instance info file)))
+      (fasl-validate-structure info file)))
 
   (let* ((2comp (component-info component))
          (entries (sb-c::ir2-component-entries 2comp))
@@ -1341,17 +1345,19 @@
 ;;;   (defvar *cpus* (... (sb-alien:alien-funcall ...)))
 ;;; and the expansion of alien-funcall involves an ALIEN-TYPE literal
 ;;; which gets multiply expanded exactly as described above.
-
 (defun load-form-is-default-mlfss-p (struct)
   ;; FIXME? this is called while writing a fasl and so might need
   ;; to invoke MAKE-LOAD-FORM, long after IR1 conversion has happened.
   ;; Surely this is not the best design.
-  (eq (nth-value 1 (sb-c::%make-load-form struct)) 'fop-struct))
+  (and (typep struct 'structure-object)
+       (multiple-value-bind (creation-form init-form)
+           (handler-case (make-load-form struct (make-null-lexenv))
+             (error (condition) (sb-c:compiler-error condition)))
+         (multiple-value-bind (ss-creation-form ss-init-form)
+             (make-load-form-saving-slots struct)
+           (and (equal creation-form ss-creation-form)
+                (equal init-form ss-init-form))))))
 
-;; Having done nothing more than load all files in obj/from-host, the
-;; cross-compiler running under any host Lisp begins life able to access
-;; SBCL-format metadata for any structure that is a subtype of STRUCTURE!OBJECT.
-;; But if it learns a layout by cross-compiling a DEFSTRUCT, that's ok too.
 (defun dump-structure (struct file)
   (unless (or (gethash struct (fasl-output-valid-structures file))
               (typep struct
@@ -1410,6 +1416,31 @@
             (logand flags sb-kernel::layout-flags-mask)
             (wrapper-length obj)))
 
+;;;; dumping instances which just save their slots
+
+(defun dump-instance-saving-slots (object slot-names file)
+  (note-potential-circularity object file)
+  (let ((circ (fasl-output-circularity-table file)))
+    (dolist (slot-name slot-names)
+      (if (slot-boundp object slot-name)
+          (let* ((value (slot-value object slot-name))
+                 (ref (gethash value circ)))
+            (cond (ref
+                   (push (make-circularity :type :slot-set
+                                           :object object
+                                           :index 0
+                                           :slot-name slot-name
+                                           :value value
+                                           :enclosing-object ref)
+                         *circularities-detected*)
+                   (sub-dump-object nil file))
+                  (t
+                   (sub-dump-object value file))))
+          (dump-fop 'fop-misc-trap file))
+      (sub-dump-object slot-name file)))
+  (sub-dump-object (class-name (class-of object)) file)
+  (dump-fop 'fop-instance file (length slot-names)))
+
 ;;;; code coverage
 
 (defun dump-code-coverage-records (namestring cc file)
