@@ -18,13 +18,11 @@
 
 ;;; Map of code-component -> list of PC offsets at which allocations occur.
 ;;; This table is needed in order to enable allocation profiling.
-(define-load-time-global *allocation-point-fixups*
+(define-load-time-global *allocation-patch-points*
   (make-hash-table :test 'eq :weakness :key :synchronized t))
 
 #-x86-64
 (progn
-(defun convert-alloc-point-fixups (dummy1 dummy2)
-  (declare (ignore dummy1 dummy2)))
 (defun sb-vm::statically-link-code-obj (code fixups)
   (declare (ignore code fixups))))
 
@@ -37,7 +35,7 @@
   (defun sb-vm::function-raw-address (name &aux (fun (fdefinition name)))
     (cond ((not (immobile-space-obj-p fun))
            (error "Can't statically link to ~S: code is movable" name))
-          ((neq (fun-subtype fun) sb-vm:simple-fun-widetag)
+          ((neq (%fun-pointer-widetag fun) sb-vm:simple-fun-widetag)
            (error "Can't statically link to ~S: non-simple function" name))
           (t
            (let ((addr (get-lisp-obj-address fun)))
@@ -57,69 +55,80 @@
 ;;; FUN must be pinned when calling this.
 (declaim (inline assign-simple-fun-self))
 (defun assign-simple-fun-self (fun)
-  (setf (%simple-fun-self fun)
-        ;; x86 backends store the address of the entrypoint in 'self'
-        #+(or x86 x86-64)
-        (%make-lisp-obj
-         (truly-the word (+ (get-lisp-obj-address fun)
-                            (ash sb-vm:simple-fun-insts-offset sb-vm:word-shift)
-                            (- sb-vm:fun-pointer-lowtag))))
-        ;; non-x86 backends store the function itself (what else?) in 'self'
-        #-(or x86 x86-64) fun))
+  (let ((self ;; x86 backends store the address of the entrypoint in 'self'
+          #+(or x86 x86-64 arm64)
+          (%make-lisp-obj
+           (truly-the word (+ (get-lisp-obj-address fun)
+                              (ash sb-vm:simple-fun-insts-offset sb-vm:word-shift)
+                              (- sb-vm:fun-pointer-lowtag))))
+          ;; non-x86 backends store the function itself (what else?) in 'self'
+          #-(or x86 x86-64 arm64) fun))
+    #-darwin-jit
+    (setf (sb-vm::%simple-fun-self fun) self)
+    #+darwin-jit
+    (sb-vm::jit-patch (+ (get-lisp-obj-address fun)
+                         (- sb-vm:fun-pointer-lowtag)
+                         (* sb-vm:simple-fun-self-slot sb-vm:n-word-bytes))
+               (get-lisp-obj-address self))))
 
-(flet ((fixup (code-obj offset sym kind flavor preserved-lists statically-link-p)
+(flet ((fixup (code-obj offset name kind flavor preserved-lists statically-link-p)
          (declare (ignorable statically-link-p))
+         ;; NAME depends on the kind and flavor of fixup.
          ;; PRESERVED-LISTS is a vector of lists of locations (by kind)
          ;; at which fixup must be re-applied after code movement.
          ;; CODE-OBJ must already be pinned in order to legally call this.
          ;; One call site that reaches here is below at MAKE-CORE-COMPONENT
          ;; and the other is LOAD-CODE, both of which pin the code.
-         ;; SYM is a little bit of a misnomer - it may be a generalized function name.
-         (when (sb-vm:fixup-code-object
+         (ecase
+             (sb-vm:fixup-code-object
                  code-obj offset
                  (ecase flavor
                    ((:assembly-routine :assembly-routine* :asm-routine-nil-offset)
-                    (- (or (get-asm-routine sym (eq flavor :assembly-routine*))
-                           (error "undefined assembler routine: ~S" sym))
+                    (- (or (get-asm-routine name (eq flavor :assembly-routine*))
+                           (error "undefined assembler routine: ~S" name))
                        (if (eq flavor :asm-routine-nil-offset) sb-vm:nil-value 0)))
-                   (:foreign (foreign-symbol-address sym))
-                   (:foreign-dataref (foreign-symbol-address sym t))
+                   (:foreign (foreign-symbol-address name))
+                   (:foreign-dataref (foreign-symbol-address name t))
                    (:code-object (get-lisp-obj-address code-obj))
-                   #+sb-thread (:symbol-tls-index (ensure-symbol-tls-index sym))
+                   #+sb-thread (:symbol-tls-index (ensure-symbol-tls-index name))
                    (:layout (get-lisp-obj-address
-                             (if (symbolp sym) (find-layout sym) sym)))
-                   (:layout-id (layout-id sym))
-                   (:immobile-symbol (get-lisp-obj-address sym))
-                   (:symbol-value (get-lisp-obj-address (symbol-global-value sym)))
+                             (wrapper-friend (if (symbolp name) (find-layout name) name))))
+                   (:layout-id (layout-id name))
+                   #+gencgc (:gc-barrier (extern-alien "gc_card_table_nbits" int))
+                   (:immobile-symbol (get-lisp-obj-address name))
+                   ;; It is legal to take the address of symbol-value only if the
+                   ;; value is known to be an immobile object
+                   ;; (whose address we don't want to wire in).
+                   (:symbol-value (get-lisp-obj-address (symbol-global-value name)))
                    #+immobile-code
                    (:named-call
-                    (prog1 (sb-vm::fdefn-entry-address sym) ; creates if didn't exist
+                    (prog1 (sb-vm::fdefn-entry-address name) ; creates if didn't exist
                       (when statically-link-p
-                        (push (cons offset (find-fdefn sym)) (elt preserved-lists 0)))))
-                   #+immobile-code (:static-call (sb-vm::function-raw-address sym)))
+                        (push (cons offset (find-fdefn name)) (elt preserved-lists 0)))))
+                   #+immobile-code (:static-call (sb-vm::function-raw-address name)))
                  kind flavor)
-           (ecase kind
-             (:relative (push offset (elt preserved-lists 1)))
-             (:absolute (push offset (elt preserved-lists 2)))
-             (:absolute64 (push offset (elt preserved-lists 3)))))
-         ;; These won't exist except for x86-64, but it doesn't matter.
-         (when (member sym '(sb-vm::enable-alloc-counter
-                             sb-vm::enable-sized-alloc-counter))
-           (push offset (elt preserved-lists 4))))
+           ((nil)) ; don't need to save it in the code-fixups, otherwise do
+           (:relative (push offset (elt preserved-lists 1)))
+           (:absolute (push offset (elt preserved-lists 2)))
+           (:immediate (push offset (elt preserved-lists 3)))
+           (:absolute64 (push offset (elt preserved-lists 4)))))
 
        (finish-fixups (code-obj preserved-lists)
          (declare (ignorable code-obj preserved-lists))
-         #+(or x86 x86-64)
+         ;; PRESERVED-LISTS are somewhat backend-dependent, but essentially
+         ;; you get to store three lists that might as well have been named
+         ;; Larry, Moe, and Curly.
          (let ((rel-fixups (elt preserved-lists 1))
                (abs-fixups (elt preserved-lists 2))
-               (abs64-fixups (elt preserved-lists 3)))
-           (aver (not abs64-fixups)) ; no preserved 64-bit fixups
-           (when (or abs-fixups rel-fixups)
+               (gc-barrier-fixups (elt preserved-lists 3))
+               (abs64-fixups (elt preserved-lists 4)))
+           ;; the fixup list packer only preserves at most 3 lists.
+           ;; And it's not clear that this is the best way to represent them.
+           (aver (not abs64-fixups))
+           (when (or abs-fixups rel-fixups gc-barrier-fixups)
              (setf (sb-vm::%code-fixups code-obj)
-                   (sb-c:pack-code-fixup-locs abs-fixups rel-fixups))))
-         (awhen (elt preserved-lists 4)
-           (setf (gethash code-obj *allocation-point-fixups*)
-                 (convert-alloc-point-fixups code-obj it)))
+                   (sb-c:pack-code-fixup-locs abs-fixups rel-fixups gc-barrier-fixups))))
+
          ;; Assign all SIMPLE-FUN-SELF slots
          (dotimes (i (code-n-entries code-obj))
            (let ((fun (%code-entry-point code-obj i)))
@@ -129,7 +138,7 @@
              (setf (sap-ref-32 (int-sap (get-lisp-obj-address fun))
                                (- 4 sb-vm:fun-pointer-lowtag))
                    (truly-the (unsigned-byte 32)
-                     (get-lisp-obj-address #.(find-layout 'function))))))
+                     (get-lisp-obj-address (wrapper-friend #.(find-layout 'function)))))))
          ;; And finally, make the memory range executable
          #-(or x86 x86-64) (sb-vm:sanctify-for-execution code-obj)
          ;; Return fixups amenable to static linking
@@ -138,11 +147,12 @@
   (defun apply-fasl-fixups (fop-stack code-obj n-fixups &aux (top (svref fop-stack 0)))
     (dx-let ((preserved (make-array 5 :initial-element nil)))
       (macrolet ((pop-fop-stack () `(prog1 (svref fop-stack top) (decf top))))
+        (binding* ((alloc-points (pop-fop-stack) :exit-if-null))
+          (setf (gethash code-obj *allocation-patch-points*) alloc-points))
         (dotimes (i n-fixups (setf (svref fop-stack 0) top))
           (multiple-value-bind (offset kind flavor)
               (sb-fasl::!unpack-fixup-info (pop-fop-stack))
-            (fixup code-obj offset (pop-fop-stack) kind flavor
-                   preserved nil))))
+            (fixup code-obj offset (pop-fop-stack) kind flavor preserved nil))))
       (finish-fixups code-obj preserved)))
 
   (defun apply-core-fixups (fixup-notes code-obj)
@@ -174,7 +184,10 @@
          (copy (allocate-code-object
                 :dynamic (code-n-named-calls code) boxed unboxed)))
     (with-pinned-objects (code copy)
+      #-darwin-jit
       (%byte-blt (code-instructions code) 0 (code-instructions copy) 0 unboxed)
+      #+darwin-jit
+      (sb-vm::jit-memcpy (code-instructions copy) (code-instructions code) unboxed)
       ;; copy boxed constants so that the fixup step (if needed) sees the 'fixups'
       ;; slot from the new object.
       (loop for i from 2 below boxed
@@ -241,15 +254,15 @@
 
 (defun assign-code-serialno (code-obj)
   (let* ((serialno (ldb (byte (byte-size sb-vm::code-serialno-byte) 0)
-                        (atomic-incf sb-fasl::*code-serialno*)))
+                        (atomic-incf *code-serialno*)))
          (insts (code-instructions code-obj))
          (jumptable-word (sap-ref-word insts 0)))
     (aver (zerop (ash jumptable-word -14)))
-    (setf (sap-ref-word insts 0) ; insert serialno
+    (setf (sb-vm::sap-ref-word-jit insts 0) ; insert serialno
           (logior (ash serialno (byte-position sb-vm::code-serialno-byte))
                   jumptable-word))))
 
-(defun make-core-component (component segment length fixup-notes object)
+(defun make-core-component (component segment length fixup-notes alloc-points object)
   (declare (type component component)
            (type segment segment)
            (type index length)
@@ -259,11 +272,25 @@
          (2comp (component-info component))
          (constants (ir2-component-constants 2comp))
          (nboxed (align-up (length constants) sb-c::code-boxed-words-align))
-         (code-obj (allocate-code-object
-                    (component-mem-space component)
-                    (count-if (lambda (x) (typep x '(cons (eql :named-call))))
-                              constants)
-                    nboxed length))
+         (n-named-calls
+          ;; Pre-scan for fdefinitions to ensure their existence.
+          ;; Doing so guarantees that storing them into the boxed header now
+          ;; can't create any old->young pointer, which is important since gencgc
+          ;; does not deal with untagged pointers when looking for old->young.
+          (do ((count 0)
+               (index (+ sb-vm:code-constants-offset
+                         (* (length (ir2-component-entries 2comp))
+                            sb-vm:code-slots-per-simple-fun))
+                      (1+ index)))
+              ((>= index (length constants)) count)
+            (let* ((const (aref constants index))
+                   (kind (if (listp const) (car const) const)))
+              (case kind
+                ((member :named-call :fdefinition)
+                 (setf (second const) (find-or-create-fdefn (second const)))
+                 (when (eq kind :named-call) (incf count)))))))
+         (code-obj (allocate-code-object (component-mem-space component)
+                                         n-named-calls nboxed length))
          (named-call-fixups
       ;; The following operations need the code pinned:
       ;; 1. copying into code-instructions (a SAP)
@@ -280,7 +307,12 @@
              ;; is GC-safe because we no longer need to know where simple-funs are embedded
              ;; within the object to trace pointers. We *do* need to know where the funs
              ;; are when transporting the object, but it's currently pinned.
+             #-darwin-jit
              (%byte-blt bytes 0 (code-instructions code-obj) 0 (length bytes))
+             #+darwin-jit
+             (with-pinned-objects (bytes)
+               (sb-vm::jit-memcpy (code-instructions code-obj) (vector-sap bytes) (length bytes)))
+
              ;; Serial# shares a word with the jump-table word count,
              ;; so we can't assign serial# until after all raw bytes are copied in.
              (assign-code-serialno code-obj))
@@ -293,8 +325,14 @@
            ;; That is, C code can't deal with an interior code pointer until the fun-table
            ;; is valid. This store must occur prior to calling %CODE-ENTRY-POINT, and
            ;; applying fixups calls %CODE-ENTRY-POINT, so we have to do this before that.
-           (setf (%code-debug-info code-obj) debug-info)
-           (apply-core-fixups fixup-notes code-obj))))
+            (setf (%code-debug-info code-obj) debug-info)
+            (apply-core-fixups fixup-notes code-obj))))
+
+    (when alloc-points
+      #+(and x86-64 sb-thread)
+      (if (= (extern-alien "alloc_profiling" int) 0) ; record the object for later
+          (setf (gethash code-obj *allocation-patch-points*) alloc-points)
+          (funcall 'sb-aprof::patch-code code-obj alloc-points)))
 
       ;; Don't need code pinned now
       ;; (It will implicitly be pinned on the conservatively scavenged backends)
@@ -304,6 +342,7 @@
         (let ((fun (%code-entry-point code-obj (decf fun-index)))
               (w (+ sb-vm:code-constants-offset
                     (* sb-vm:code-slots-per-simple-fun fun-index))))
+          (aver (functionp fun)) ; in case %CODE-ENTRY-POINT returns NIL
           (setf (code-header-ref code-obj (+ w sb-vm:simple-fun-name-slot))
                 (entry-info-name entry-info)
                 (code-header-ref code-obj (+ w sb-vm:simple-fun-arglist-slot))
@@ -330,8 +369,7 @@
           (t
            (let ((referent
                   (etypecase kind
-                   ((member :named-call :fdefinition)
-                    (find-or-create-fdefn (cadr const)))
+                   ((member :named-call :fdefinition) (cadr const))
                    ((eql :known-fun)
                     (%coerce-name-to-fun (cadr const)))
                    (constant
@@ -341,6 +379,9 @@
                  (setf (code-header-ref code-obj index) referent)))))))
     (when named-call-fixups
       (sb-vm::statically-link-code-obj code-obj named-call-fixups))
+    (when sb-fasl::*show-new-code*
+      (let ((*print-pretty* nil))
+        (format t "~&New code(~Db,core): ~A~%" (code-object-size code-obj) code-obj)))
     code-obj))
 
 (defun set-code-fdefn (code index fdefn)
