@@ -16,41 +16,75 @@
 
 ;;; PPC64 can't use the NIL-as-CONS + NIL-as-symbol trick *and* avoid using
 ;;; temp-reg-tn to access symbol slots.
-;;; Since the NIL-as-CONS is necessary, and efficient accessor to lists and
+;;; Since the NIL-as-CONS is necessary, and efficient access to lists and
 ;;; instances is desirable, we lose a little on symbol access by being forced
-;;; to pre-check for NIL. There is trick that can get back some performance
-;;; on SYMBOL-VALUE which I plan to implement after this much works right.
+;;; to pre-check for NIL.
+
+(defun read-symbol-slot (slot symbol result)
+  (let ((null-label (gen-label))
+        (done-label (gen-label)))
+    (inst cmpld symbol null-tn)
+    (inst beq null-label)
+    (loadw result symbol slot other-pointer-lowtag)
+    (inst b done-label)
+    (emit-label null-label)
+    ;; This is a very un-memorable bit of fudge factor magic
+    ;; that I have to re-figure-out every time I need it.
+    ;; Some sort of abstraction might be nice. Or not.
+    (loadw result symbol (1- slot) list-pointer-lowtag)
+    (emit-label done-label)))
+
 (define-vop (slot)
   (:args (object :scs (descriptor-reg)))
   (:info name offset lowtag)
   (:results (result :scs (descriptor-reg any-reg)))
   (:generator 1
-    (cond ((member name '(symbol-name symbol-info sb-xc:symbol-package))
-           (let ((null-label (gen-label))
-                 (done-label (gen-label)))
-             (inst cmpld object null-tn)
-             (inst beq null-label)
-             (loadw result object offset lowtag)
-             (inst b done-label)
-             (emit-label null-label)
-             (loadw result object (1- offset) list-pointer-lowtag)
-             (emit-label done-label)))
+    (cond ((member name '(symbol-%info %symbol-fdefn))
+           (read-symbol-slot offset object result))
           (t
            (loadw result object offset lowtag)))))
 
+(define-vop ()
+  (:args (symbol :scs (descriptor-reg)))
+  (:results (result :scs (unsigned-reg)))
+  (:result-types positive-fixnum)
+  (:translate sb-impl::symbol-package-id)
+  (:policy :fast-safe)
+  (:generator 5
+   (inst cmpld symbol null-tn)
+   ;; This loads "random" bits if SYMBOL is NIL, but since RESULT is non-descriptor
+   ;; there is no harm. We check for NULL and move a constant to the result if so.
+   ;; ASSUMPTION: symbol-package-bits = 16
+   (inst lhz result symbol (+ (ash symbol-name-slot word-shift)
+                              (- other-pointer-lowtag)
+                              #+little-endian 6))
+   (inst bne done)
+   (inst lr result 1) ; SB-IMPL::+PACKAGE-ID-LISP+
+   DONE))
+(define-vop ()
+  (:args (symbol :scs (descriptor-reg)))
+  (:results (result :scs (descriptor-reg)))
+  (:translate symbol-name)
+  (:policy :fast-safe)
+  (:temporary (:sc non-descriptor-reg :offset nl3-offset) pa-flag)
+  (:generator 2
+   (pseudo-atomic (pa-flag :sync nil)
+     (read-symbol-slot symbol-name-slot symbol result)
+     (inst rldicl result result 0 sb-impl::package-id-bits))))
+
 (define-vop (set-slot)
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg)))
   (:info name offset lowtag)
-  (:ignore name)
-  (:results)
+  (:temporary (:scs (non-descriptor-reg)) t1)
+  (:vop-var vop)
   (:generator 1
+    ;; gencgc does not need to emit the barrier for constructors
+    (unless (member name '(%make-structure-instance make-weak-pointer
+                           %make-ratio %make-complex))
+      (emit-gc-store-barrier object nil (list t1) (vop-nth-arg 1 vop) value))
     (storew value object offset lowtag)))
 
-(define-vop (init-slot set-slot)
-  (:info name dx-p offset lowtag)
-  (:ignore name dx-p))
-
 (define-vop (compare-and-swap-slot)
   (:args (object :scs (descriptor-reg))
          (old :scs (descriptor-reg any-reg))
@@ -59,7 +93,9 @@
   (:info name offset lowtag)
   (:ignore name)
   (:results (result :scs (descriptor-reg) :from :load))
+  (:vop-var vop)
   (:generator 5
+    (emit-gc-store-barrier object nil (list temp) (vop-nth-arg 2 vop) new)
     (inst sync)
     (inst li temp (- (* offset n-word-bytes) lowtag))
     LOOP
@@ -84,19 +120,17 @@
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 15
+    (emit-gc-store-barrier symbol nil (list temp) (vop-nth-arg 2 vop) new)
     (inst sync)
-    #+sb-thread
-    (assemble ()
-      (load-tls-index temp symbol)
-      ;; Thread-local area, no synchronization needed.
-      (inst ldx result thread-base-tn temp)
-      (inst cmpd result old)
-      (inst bne DONT-STORE-TLS)
-      (inst stdx new thread-base-tn temp)
-      DONT-STORE-TLS
-
-      (inst cmpdi result no-tls-value-marker-widetag)
-      (inst bne CHECK-UNBOUND))
+    (load-tls-index temp symbol)
+    ;; Thread-local area, no synchronization needed.
+    (inst ldx result thread-base-tn temp)
+    (inst cmpd result old)
+    (inst bne DONT-STORE-TLS)
+    (inst stdx new thread-base-tn temp)
+    DONT-STORE-TLS
+    (inst cmpdi result no-tls-value-marker-widetag)
+    (inst bne CHECK-UNBOUND)
 
     (inst li temp (- (* symbol-value-slot n-word-bytes)
                      other-pointer-lowtag))
@@ -126,7 +160,6 @@
   (:temporary (:scs (descriptor-reg) :from (:argument 0)) obj-temp))
 
 ;;; With SYMBOL-VALUE, we check that the value isn't the trap object.
-;;; So SYMBOL-VALUE of NIL is NIL.
 (define-vop (symbol-global-value checked-cell-ref)
   (:translate sym-global-val)
   (:generator 9
@@ -157,89 +190,77 @@
     (move value object)
     DONE))
 
-#+sb-thread
-(progn
-  (define-vop (set)
-    (:args (symbol :scs (descriptor-reg))
-           (value :scs (descriptor-reg any-reg)))
-    (:temporary (:sc any-reg) tls-slot temp)
-    (:generator 4
-      (load-tls-index tls-slot symbol)
-      (inst ldx temp thread-base-tn tls-slot)
-      (inst cmpdi temp no-tls-value-marker-widetag)
-      (inst beq GLOBAL-VALUE)
-      (inst stdx value thread-base-tn tls-slot)
-      (inst b DONE)
-      GLOBAL-VALUE
-      (storew value symbol symbol-value-slot other-pointer-lowtag)
-      DONE))
+(define-vop (set)
+  (:args (symbol :scs (descriptor-reg))
+         (value :scs (descriptor-reg any-reg)))
+  (:temporary (:sc non-descriptor-reg) tls-slot)
+  (:temporary (:sc any-reg) temp)
+  (:vop-var vop)
+  (:generator 4
+    (load-tls-index tls-slot symbol)
+    (inst ldx temp thread-base-tn tls-slot)
+    (inst cmpdi temp no-tls-value-marker-widetag)
+    (inst beq GLOBAL-VALUE)
+    (inst stdx value thread-base-tn tls-slot)
+    (inst b DONE)
+    GLOBAL-VALUE
+    (emit-gc-store-barrier symbol nil (list tls-slot) (vop-nth-arg 1 vop) value)
+    (storew value symbol symbol-value-slot other-pointer-lowtag)
+    DONE))
 
-  ;; With Symbol-Value, we check that the value isn't the trap object. So
-  ;; Symbol-Value of NIL is NIL.
-  (define-vop (symbol-value)
-    (:translate symeval)
-    (:policy :fast-safe)
-    (:args (object :scs (descriptor-reg) :to (:result 1)))
-    (:results (value :scs (descriptor-reg any-reg)))
-    (:vop-var vop)
-    (:save-p :compute-only)
-    (:generator 9
-      (inst cmpld object null-tn)
-      (inst beq NULL)
-      (load-tls-index value object)
-      (inst ldx value thread-base-tn value)
-      (inst cmpdi value no-tls-value-marker-widetag)
-      (inst bne CHECK-UNBOUND)
-      (loadw value object symbol-value-slot other-pointer-lowtag)
-      CHECK-UNBOUND
-      (inst cmpdi value unbound-marker-widetag)
-      (inst beq (generate-error-code vop 'unbound-symbol-error object))
-      (inst b DONE)
-      NULL
-      (move value object)
-      DONE))
+;; With Symbol-Value, we check that the value isn't the trap object.
+(define-vop (symbol-value)
+  (:translate symeval)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg) :to (:result 1)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:vop-var vop)
+  (:save-p :compute-only)
+  (:generator 9
+    (inst cmpld object null-tn)
+    (inst beq NULL)
+    (load-tls-index value object)
+    (inst ldx value thread-base-tn value)
+    (inst cmpdi value no-tls-value-marker-widetag)
+    (inst bne CHECK-UNBOUND)
+    (loadw value object symbol-value-slot other-pointer-lowtag)
+    CHECK-UNBOUND
+    (inst cmpdi value unbound-marker-widetag)
+    (inst beq (generate-error-code vop 'unbound-symbol-error object))
+    (inst b DONE)
+    NULL
+    (move value object)
+    DONE))
 
-  (define-vop (fast-symbol-value symbol-value)
-    ;; KLUDGE: not really fast, in fact, because we're going to have to
-    ;; do a full lookup of the thread-local area anyway.  But half of
-    ;; the meaning of FAST-SYMBOL-VALUE is "do not signal an error if
-    ;; unbound", which is used in the implementation of COPY-SYMBOL.  --
-    ;; CSR, 2003-04-22
-    (:policy :fast)
-    (:translate symeval)
-    (:generator 8
-      (inst cmpld object null-tn)
-      (inst beq NULL)
-      (load-tls-index value object)
-      (inst ldx value thread-base-tn value)
-      (inst cmpdi value no-tls-value-marker-widetag)
-      (inst bne DONE)
-      (loadw value object symbol-value-slot other-pointer-lowtag)
-      (inst b DONE)
-      NULL
-      (move value object)
-      DONE)))
-
-;;; On unithreaded builds these are just copies of the global versions.
-#-sb-thread
-(progn
-  (define-vop (symbol-value symbol-global-value)
-    (:translate symeval))
-  (define-vop (fast-symbol-value fast-symbol-global-value)
-    (:translate symeval))
-  (define-vop (set %set-symbol-global-value)))
+(define-vop (fast-symbol-value symbol-value)
+  ;; KLUDGE: not really fast, in fact, because we're going to have to
+  ;; do a full lookup of the thread-local area anyway.  But half of
+  ;; the meaning of FAST-SYMBOL-VALUE is "do not signal an error if
+  ;; unbound", which is used in the implementation of COPY-SYMBOL.  --
+  ;; CSR, 2003-04-22
+  (:policy :fast)
+  (:translate symeval)
+  (:generator 8
+    (inst cmpld object null-tn)
+    (inst beq NULL)
+    (load-tls-index value object)
+    (inst ldx value thread-base-tn value)
+    (inst cmpdi value no-tls-value-marker-widetag)
+    (inst bne DONE)
+    (loadw value object symbol-value-slot other-pointer-lowtag)
+    (inst b DONE)
+    NULL
+    (move value object)
+    DONE))
 
 ;;; Like CHECKED-CELL-REF, only we are a predicate to see if the cell
 ;;; is bound.
-(define-vop (boundp-frob)
+(define-vop (boundp)
   (:args (object :scs (descriptor-reg)))
   (:conditional)
   (:info target not-p)
   (:policy :fast-safe)
-  (:temporary (:scs (descriptor-reg)) value))
-
-#+sb-thread
-(define-vop (boundp boundp-frob)
+  (:temporary (:scs (descriptor-reg)) value)
   (:translate boundp)
   (:generator 9
     (inst cmpld object null-tn)
@@ -254,14 +275,6 @@
     (inst b? (if not-p :eq :ne) target)
     OUT))
 
-#-sb-thread
-(define-vop (boundp boundp-frob)
-  (:translate boundp)
-  (:generator 9
-    (loadw value object symbol-value-slot other-pointer-lowtag)
-    (inst cmpwi value unbound-marker-widetag)
-    (inst b? (if not-p :eq :ne) target)))
-
 (define-vop (symbol-hash)
   (:policy :fast-safe)
   (:translate symbol-hash)
@@ -280,27 +293,6 @@
     NULL
     (inst addi res null-tn (- (logand sb-vm:nil-value sb-vm:fixnum-tag-mask)))
     DONE))
-(define-vop (symbol-plist)
-  (:policy :fast-safe)
-  (:translate symbol-plist)
-  (:args (symbol :scs (descriptor-reg)))
-  (:results (res :scs (descriptor-reg)))
-  (:temporary (:scs (unsigned-reg)) temp)
-  (:generator 6
-    (inst cmpld symbol null-tn)
-    (inst beq NULL)
-    (loadw res symbol symbol-info-slot other-pointer-lowtag)
-    (inst andi. temp res lowtag-mask)
-    (inst cmpwi temp list-pointer-lowtag)
-    (inst beq take-car)
-    (move res null-tn) ; if INFO is a non-list, then the PLIST is NIL
-    (inst b DONE)
-    NULL
-    (loadw res symbol (1- symbol-info-slot) list-pointer-lowtag)
-    ;; fallthru. NULL's info slot always holds a cons
-    TAKE-CAR
-    (loadw res res cons-car-slot list-pointer-lowtag)
-    DONE))
 
 ;;;; Fdefinition (fdefn) objects.
 
@@ -358,6 +350,7 @@
   (:temporary (:scs (non-descriptor-reg)) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
+    (emit-gc-store-barrier fdefn nil (list type))
     (let ((normal-fn (gen-label)))
       (load-type type function (- fun-pointer-lowtag))
       (inst cmpdi type simple-fun-widetag)
@@ -374,16 +367,12 @@
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
+  (:args (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
     (inst addi temp null-tn (make-fixup 'undefined-tramp :asm-routine-nil-offset))
-    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (move result fdefn)))
-
-
+    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 ;;;; Binding and Unbinding.
 
@@ -391,7 +380,6 @@
 ;;; the symbol on the binding stack and stuff the new value into the
 ;;; symbol.
 ;;; See the "Chapter 9: Specials" of the SBCL Internals Manual.
-#+sb-thread
 (define-vop (dynbind)
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
@@ -406,19 +394,6 @@
      (storew tls-index bsp-tn (- binding-symbol-slot binding-size))
      (inst stdx val thread-base-tn tls-index))))
 
-#-sb-thread
-(define-vop (dynbind)
-  (:args (val :scs (any-reg descriptor-reg))
-         (symbol :scs (descriptor-reg)))
-  (:temporary (:scs (descriptor-reg)) temp)
-  (:generator 5
-    (loadw temp symbol symbol-value-slot other-pointer-lowtag)
-    (inst addi bsp-tn bsp-tn (* binding-size n-word-bytes))
-    (storew temp bsp-tn (- binding-value-slot binding-size))
-    (storew symbol bsp-tn (- binding-symbol-slot binding-size))
-    (storew val symbol symbol-value-slot other-pointer-lowtag)))
-
-#+sb-thread
 (define-vop (unbind)
   (:temporary (:scs (descriptor-reg)) tls-index value)
   (:temporary (:scs (any-reg)) zero)
@@ -431,20 +406,6 @@
     (storew zero bsp-tn (- binding-value-slot binding-size))
     (inst subi bsp-tn bsp-tn (* binding-size n-word-bytes))))
 
-#-sb-thread
-(define-vop (unbind)
-  (:temporary (:scs (descriptor-reg)) symbol value)
-  (:temporary (:scs (any-reg)) zero)
-  (:generator 0
-    (loadw symbol bsp-tn (- binding-symbol-slot binding-size))
-    (loadw value bsp-tn (- binding-value-slot binding-size))
-    (storew value symbol symbol-value-slot other-pointer-lowtag)
-    (inst li zero 0)
-    (storew zero bsp-tn (- binding-symbol-slot binding-size))
-    (storew zero bsp-tn (- binding-value-slot binding-size))
-    (inst subi bsp-tn bsp-tn (* binding-size n-word-bytes))))
-
-
 (define-vop (unbind-to-here)
   (:args (arg :scs (descriptor-reg any-reg) :target where))
   (:temporary (:scs (any-reg) :from (:argument 0)) where zero)
@@ -460,10 +421,7 @@
       (inst cmpdi symbol 0)
       (inst beq skip)
       (loadw value bsp-tn (- binding-value-slot binding-size))
-      #+sb-thread
       (inst stdx value thread-base-tn symbol)
-      #-sb-thread
-      (storew value symbol symbol-value-slot other-pointer-lowtag)
       (storew zero bsp-tn (- binding-symbol-slot binding-size))
 
       SKIP
@@ -482,14 +440,14 @@
   (:variant closure-info-offset fun-pointer-lowtag)
   (:translate %closure-index-ref))
 
+(define-vop (%closure-index-set descriptor-word-index-set)
+  (:variant closure-info-offset fun-pointer-lowtag)
+  (:translate %closure-index-set))
+
 (define-vop (funcallable-instance-info word-index-ref)
   (:variant funcallable-instance-info-offset fun-pointer-lowtag)
   (:translate %funcallable-instance-info))
 
-(define-vop (set-funcallable-instance-info word-index-set-nr)
-  (:variant funcallable-instance-info-offset fun-pointer-lowtag)
-  (:translate %set-funcallable-instance-info))
-
 (define-vop (closure-ref)
   (:args (object :scs (descriptor-reg)))
   (:results (value :scs (descriptor-reg any-reg)))
@@ -538,7 +496,7 @@
   (:variant instance-slots-offset instance-pointer-lowtag)
   (:arg-types instance positive-fixnum))
 
-(define-vop (instance-index-set word-index-set)
+(define-vop (instance-index-set descriptor-word-index-set)
   (:policy :fast-safe)
   (:translate %instance-set)
   (:variant instance-slots-offset instance-pointer-lowtag)
@@ -562,31 +520,22 @@
 
 ;;;; Code object frobbing.
 
-(define-vop (code-header-ref-any)
+(define-vop (code-header-ref+tag)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg)))
   (:arg-types * tagged-num)
+  (:info implied-lowtag)
+  ;; conservative_root_p() in gencgc treats untagged pointers to fdefns
+  ;; as implicitly pinned. It has to be in a boxed register.
   (:results (value :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 2
     ;; ASSUMPTION: N-FIXNUM-TAG-BITS = 3
     (inst addi temp index (- other-pointer-lowtag))
-    (inst ldx value object temp)))
-
-(define-vop (code-header-ref-fdefn)
-  (:args (object :scs (descriptor-reg))
-         (index :scs (any-reg)))
-  (:arg-types * tagged-num)
-  (:results (value :scs (descriptor-reg)))
-  (:policy :fast-safe)
-  (:temporary (:scs (non-descriptor-reg)) temp)
-  (:generator 3
-    ;; ASSUMPTION: N-FIXNUM-TAG-BITS = 3
-    (inst addi temp index (- other-pointer-lowtag))
-    ;; Loaded value is automatically pinned.
     (inst ldx value object temp)
-    (inst ori value value other-pointer-lowtag)))
+    (unless (zerop implied-lowtag)
+      (inst ori value value implied-lowtag))))
 
 #-sb-xc-host
 (defun code-header-ref (code index)
@@ -594,16 +543,36 @@
   (let ((fdefns-start (sb-impl::code-fdefns-start-index code))
         (count (code-n-named-calls code)))
     (declare ((unsigned-byte 16) fdefns-start count))
+    (values
     (if (and (>= index fdefns-start) (< index (+ fdefns-start count)))
-        (%primitive code-header-ref-fdefn code index)
-        (%primitive code-header-ref-any code index))))
+        (%primitive code-header-ref+tag code index other-pointer-lowtag)
+        (%primitive code-header-ref+tag code index 0)))))
 
-(define-vop (code-header-set word-index-set-nr)
+(define-vop (code-header-set)
   (:translate code-header-set)
   (:policy :fast-safe)
-  (:variant 0 other-pointer-lowtag))
-
-
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * tagged-num *)
+  (:temporary (:scs (non-descriptor-reg)) temp card)
+  (:temporary (:sc non-descriptor-reg :offset nl3-offset) pa-flag)
+  (:generator 10
+    ;; Load mark table base
+    (inst ld temp thread-base-tn (ash thread-card-table-slot word-shift))
+    (pseudo-atomic (pa-flag)
+      ;; Compute card mark index
+      (inst rldicl card object (- 64 gencgc-card-shift) (make-fixup nil :gc-barrier))
+      ;; Touch the card mark byte.
+      (inst stbx thread-base-tn temp card) ; THREAD-TN's low byte is 0
+      ;; set 'written' flag in the code header
+      ;; If two threads get here at the same time, they'll write the same byte.
+      (let ((byte (- #+big-endian 4 #+little-endian 3 other-pointer-lowtag)))
+        (inst lbz temp object byte)
+        (inst ori temp temp #x40)
+        (inst stb temp object byte))
+      (inst addi temp index (- other-pointer-lowtag))
+      (inst stdx value object temp))))
 
 ;;;; raw instance slot accessors
 
@@ -621,7 +590,7 @@
                   (:arg-types instance positive-fixnum)
                   (:results (value :scs (,sc)))
                   (:result-types ,primtype))
-                (define-vop (,(symbolicate "%RAW-INSTANCE-SET/" suffix) word-index-set-nr)
+                (define-vop (,(symbolicate "%RAW-INSTANCE-SET/" suffix) word-index-set)
                   (:policy :fast-safe)
                   (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
                   (:variant instance-slots-offset instance-pointer-lowtag)
