@@ -214,8 +214,10 @@
 ;;; ignore TYPE-CHECK-ERROR because we don't want the possibility of
 ;;; error to bias the result. Notes are suppressed for T-C-E as well,
 ;;; since we don't need to worry about the efficiency of that case.
-(define-load-time-global *ignore-cost-vops* '(set type-check-error))
-(define-load-time-global *suppress-note-vops* '(type-check-error))
+(defconstant-eqx ignore-cost-vops '(set type-check-error) #'equal)
+(defconstant-eqx suppress-note-vops '(type-check-error) #'equal)
+
+(declaim (start-block select-tn-representation))
 
 ;;; We special-case the move VOP, since using this costs for the
 ;;; normal MOVE would spuriously encourage descriptor representations.
@@ -239,7 +241,7 @@
       (let* ((vop (tn-ref-vop ref))
              (info (vop-info vop)))
         (unless (and (neq (tn-kind tn) :constant)
-                     (find (vop-info-name info) *ignore-cost-vops*))
+                     (find (vop-info-name info) ignore-cost-vops))
           (case (vop-info-name info)
             (move
              (let ((rep (tn-sc
@@ -292,7 +294,7 @@
                             t)
 
 
-  (let ((min sb-xc:most-positive-fixnum)
+  (let ((min most-positive-fixnum)
         (min-scn nil)
         (unique nil))
     (dolist (scn scs)
@@ -305,6 +307,8 @@
                (setq unique t)))))
     (values min-scn unique)))
 
+(declaim (end-block))
+
 ;;; Prepare for the possibility of a TN being allocated on the number
 ;;; stack by setting NUMBER-STACK-P in all functions that TN is
 ;;; referenced in and in all the functions in their tail sets. REFS is
@@ -319,9 +323,9 @@
                      (vop-block (tn-ref-vop ref)))))
            (tails (lambda-tail-set lambda)))
       (flet ((frob (fun)
-               (setf (ir2-physenv-number-stack-p
-                      (physenv-info
-                       (lambda-physenv fun)))
+               (setf (ir2-environment-number-stack-p
+                      (environment-info
+                       (lambda-environment fun)))
                      t)))
         (frob lambda)
         (when tails
@@ -359,7 +363,7 @@
     (cond ((eq (tn-kind op-tn) :constant))
           ((policy op-node (and (<= speed inhibit-warnings)
                                 (<= space inhibit-warnings))))
-          ((member (template-name (vop-info op-vop)) *suppress-note-vops*))
+          ((member (template-name (vop-info op-vop)) suppress-note-vops))
           ((null dest-tn)
            (let* ((op-info (vop-info op-vop))
                   (op-note (or (template-note op-info)
@@ -439,7 +443,7 @@
 ;;;
 ;;; If the TN is an unused result TN, then we don't actually emit the
 ;;; move; we just change to the right kind of TN.
-(defun emit-coerce-vop (op dest-tn scs before)
+(defun emit-coerce-vop (op dest-tn scs before &optional load-scs)
   (declare (type tn-ref op) (type sc-vector scs) (type (or vop null) before)
            (type (or tn null) dest-tn))
   (let* ((op-tn (tn-ref-tn op))
@@ -463,7 +467,7 @@
                        (cond
                          ((not write-p)
                           (or
-                           (coerce-from-constant op temp)
+                           (coerce-from-constant op temp load-scs)
                            (emit-move (or (maybe-move-from-fixnum+-1 op-tn temp
                                                                      op)
                                           res)
@@ -495,26 +499,38 @@
 ;;; specified VOP. Dest-TN is the destination TN if we are doing a
 ;;; move or move-arg, and is NIL otherwise. This is only used for
 ;;; efficiency notes.
-(defun coerce-some-operands (ops dest-tn load-scs before)
+(defun coerce-some-operands (ops dest-tn load-scs before
+                             &optional more-load-scs)
   (declare (type (or tn-ref null) ops) (list load-scs)
            (type (or tn null) dest-tn) (type (or vop null) before))
   (do ((op ops (tn-ref-across op))
        (scs load-scs (cdr scs)))
-      ((null scs))
-    (let ((tn (tn-ref-tn op)))
+      ((null op))
+
+    (let* (more
+           (scs (cond (scs
+                       (car scs))
+                      (more-load-scs
+                       (setf more t)
+                       more-load-scs)
+                      (t
+                       (return))))
+           (tn (tn-ref-tn op)))
       (unless (or (eq (tn-kind tn) :unused)
-                  (svref (car scs)
+                  (svref scs
                          (sc-number (tn-sc tn))))
-        (emit-coerce-vop op dest-tn (car scs) before))))
+        (emit-coerce-vop op dest-tn scs before (and more
+                                                    more-load-scs)))))
   (values))
 
 ;;; Emit coerce VOPs for the args and results, as needed.
 (defun coerce-vop-operands (vop)
   (declare (type vop vop))
   (let ((info (vop-info vop)))
-    (coerce-some-operands (vop-args vop) nil (vop-info-arg-load-scs info) vop)
+    (coerce-some-operands (vop-args vop) nil (vop-info-arg-load-scs info) vop
+                          (vop-info-more-arg-load-scs info))
     (coerce-some-operands (vop-results vop) nil (vop-info-result-load-scs info)
-                          (vop-next vop)))
+                          (vop-next vop) nil))
   (values))
 
 ;;; Iterate over the more operands to a call VOP, emitting move-arg
@@ -572,8 +588,7 @@
                                             pass-tn vop))
                (after
                 (cond ((eq how :local-call)
-                       (aver (eq (vop-info-name (vop-info prev))
-                                 'allocate-frame))
+                       (aver (eq (vop-name prev) 'allocate-frame))
                        prev)
                       (prev (vop-next prev))
                       (t
@@ -589,23 +604,32 @@
     (let ((type (tn-ref-type x-tn-ref)))
       (cond ((not type)
              nil)
+            ((csubtypep type (specifier-type 'fixnum))
+             (template-or-lose 'sb-vm::move-from-word/fixnum))
             ((csubtypep type
-                        (specifier-type `(integer ,sb-xc:most-negative-fixnum
-                                                  ,(1+ sb-xc:most-positive-fixnum))))
+                        (specifier-type `(integer ,most-negative-fixnum
+                                                  ,(1+ most-positive-fixnum))))
              (template-or-lose 'sb-vm::move-from-fixnum+1))
             ((csubtypep type
-                        (specifier-type `(integer ,(1- sb-xc:most-negative-fixnum)
-                                                  ,sb-xc:most-positive-fixnum)))
+                        (specifier-type `(integer ,(1- most-negative-fixnum)
+                                                  ,most-positive-fixnum)))
              (template-or-lose 'sb-vm::move-from-fixnum-1))))))
 
-(defun coerce-from-constant (x-tn-ref y)
+(defun coerce-from-constant (x-tn-ref y &optional load-scs)
   (when (and (sc-is y sb-vm::descriptor-reg sb-vm::control-stack)
              (tn-ref-type x-tn-ref))
     (multiple-value-bind (constantp value) (type-singleton-p (tn-ref-type x-tn-ref))
       (when constantp
-        (change-tn-ref-tn x-tn-ref
-                          (make-constant-tn (find-constant value) t))
-        t))))
+        (let* ((constant (find-constant value))
+               (sc (constant-sc constant)))
+          (when (or (not load-scs)
+                    ;; This is a more-arg-load-scs
+                    ;; Because more args do not have a generic load
+                    ;; sequence it can't handle aribtrary constants or
+                    ;; immediates.
+                    (svref load-scs (sc-number sc)))
+            (change-tn-ref-tn x-tn-ref (make-constant-tn constant t))
+            t))))))
 
 (defun split-ir2-block (vop)
   (cond ((vop-next vop)
@@ -614,9 +638,11 @@
                   (succ (ir2-block-block 2block))
                   (start (make-ctran))
                   (block (ctran-starts-block start))
-                  (no-op-node (make-no-op))
+                  (no-op-node (make-exit))
                   (new-2block (make-ir2-block block))
                   (vop-next (vop-next vop)))
+             (setf (block-number block)
+                   (incf (component-max-block-number (block-component block))))
              (link-node-to-previous-ctran no-op-node start)
              (setf (block-info block) new-2block)
              (add-to-emit-order new-2block (ir2-block-prev 2block))
@@ -728,12 +754,14 @@
         ((vop-info-move-args info)
          (emit-arg-moves vop))
         (t
-         (coerce-vop-operands vop))))))
+         (coerce-vop-operands vop)))
+      (when (vop-info-after-sc-selection info)
+        (funcall (vop-info-after-sc-selection info) vop)))))
 
 ;;; If TN is in a number stack SC, make all the right annotations.
 ;;; Note that this should be called after TN has been referenced,
 ;;; since it must iterate over the referencing environments.
-#-sb-fluid (declaim (inline note-if-number-stack))
+(declaim (inline note-if-number-stack))
 (defun note-if-number-stack (tn 2comp restricted)
   (declare (type tn tn) (type ir2-component 2comp))
   (when (if restricted
@@ -745,6 +773,106 @@
     (note-number-stack-tn (tn-writes tn)))
   (values))
 
+;;; Arrange boxed constants so that all :NAMED-CALL constants are first,
+;;; then constant leaves, and finally LOAD-TIME-VALUE constants.
+;;; There exist a few reasons for placing all the FDEFNs first:
+;;;  * FDEFNs which are referenced for lisp call - as opposed to referenced
+;;;    in #'FUN syntax - could be stored as untagged pointers which would
+;;;    benefit the PPC64 architecture by removing a few instructions from each
+;;;    use of such fdefn by not having to subtract its lowtag prior to loading
+;;;    from both the fun and raw-fun slots. GC would need to be aware of the
+;;;    untagged pointer convention.
+;;;  * In the current approach for so-called "static" linking of immobile code,
+;;;    we change code instruction bytes so that they call into a simple-fun
+;;;    directly rather than through an fdefn, but the approach is subject to a
+;;;    data race when redefining an fdefn. It's conceivable that the race can be
+;;;    eliminated by substituting placeholders in the code headers of functions
+;;;    that had static linking performed - so that they see a reference to the
+;;;    callee rather than an fdefn - but in order for that to work, we must
+;;;    distinguish between fdefns that are needed for FDEFN-FUN
+;;;    (via IR2-CONVERT-GLOBAL-VAR) versus those which are present to satisfy
+;;;    a GC invariant and are not otherwise actually used.
+;;;  * Even without the preceding change, remove-static-links can avoid
+;;;    scanning code constants that are not FDEFNs.
+(defun sort-boxed-constants (2comp)
+  (let* ((sorted (ir2-component-constants 2comp))
+         (unsorted (subseq sorted 1))
+         (renumbering)) ; alist of (old . new) indices into constant vector
+    (setf (fill-pointer sorted) 0)
+    ;; add in fixed overhead
+    (let ((n-entries (length (ir2-component-entries 2comp))))
+      (dotimes (i (+ (* sb-vm:code-slots-per-simple-fun n-entries)
+                     sb-vm:code-constants-offset))
+        (vector-push-extend nil sorted)))
+    (flet ((scan (pass &aux (old-offset 0))
+             (dovector (constant unsorted)
+               (incf old-offset)
+               (when (eql pass (cond ((constant-p constant) 2)
+                                     ((eq (car constant) :named-call) 1)
+                                     (t 3)))
+                 (let ((new-offset (vector-push-extend constant sorted)))
+                   (push (cons old-offset new-offset) renumbering))))))
+      (scan 1)  ; first all the called fdefinitions
+      (scan 2)  ; then IR1 constants
+      (scan 3)) ; then various flavors of load-time magic
+    ;; Update the TN-OFFSET slot.
+    ;; There can be more than one TN with the same index into the constants
+    ;; because of how MAKE-LOAD-TIME-CONSTANT-TN works.
+    (do ((tn (ir2-component-constant-tns 2comp) (tn-next tn)))
+        ((null tn))
+      (let ((old-offset (tn-offset tn)))
+        (when old-offset
+          (setf (tn-offset tn)
+                (cdr (the (not null) (assoc old-offset renumbering)))))))
+    sorted))
+
+#+arm64
+(defun choose-zero-tn (tn)
+  (let (zero-tn)
+    (flet ((zero-tn ()
+             (or zero-tn
+                 (setf zero-tn
+                       (component-live-tn
+                        (make-wired-tn nil
+                                       sb-vm:any-reg-sc-number
+                                       sb-vm::zr-offset))))))
+      (do ((tn tn (tn-next tn)))
+          ((null tn))
+        (when (and (constant-tn-p tn)
+                   (eql (tn-value tn) 0))
+          (loop with next
+                for read = (tn-reads tn) then next
+                while read
+                do
+                (setf next (tn-ref-next read))
+                (do* ((vop (tn-ref-vop read))
+                      (info (vop-info vop))
+                      (cost (vop-info-arg-costs info) (cdr cost))
+                      (op (vop-args vop) (tn-ref-across op)))
+                     ((null cost))
+                  (when (eq op read)
+                    (when (eql (svref (car cost) sb-vm:zero-sc-number) 0)
+                      (change-tn-ref-tn read (zero-tn)))
+                    (return)))))))))
+
+;;; The call VOPs allocate a temporary register wired to
+;;; nfp-save-offset, but don't use it if there's no nfp-tn in the
+;;; current frame, but the stack space is still allocated.
+#-c-stack-is-control-stack
+(defun unwire-nfp-save-tn (2comp)
+  (unless (ir2-component-nfp 2comp)
+    (do ((prev)
+         (tn (ir2-component-wired-tns 2comp) (tn-next tn)))
+        ((null tn))
+      (cond ((and (sc-is tn sb-vm::control-stack)
+                  (eql (tn-offset tn) sb-vm:nfp-save-offset))
+             (setf (tn-kind tn) :unused)
+             (if prev
+                 (setf (tn-next prev) (tn-next tn))
+                 (setf (ir2-component-wired-tns 2comp) (tn-next tn))))
+            (t
+             (setf prev tn))))))
+
 ;;; This is the entry to representation selection. First we select the
 ;;; representation for all normal TNs, setting the TN-SC. After
 ;;; selecting the TN representations, we set the SC for all :ALIAS TNs
@@ -757,12 +885,13 @@
 (defun select-representations (component)
   (let ((costs (make-array sb-vm:sc-number-limit))
         (2comp (component-info component)))
+    (sort-boxed-constants 2comp)
     (labels ((set-sc (tn sc)
-               (cond ((not (eq (tn-kind tn) :constant))
+               (cond ((neq (tn-kind tn) :constant)
                       (setf (tn-sc tn)
                             (svref *backend-sc-numbers* sc)))
                      ;; Translate primitive type scs into constant scs
-                     ((= sc sb-vm::descriptor-reg-sc-number)
+                     ((= sc sb-vm:descriptor-reg-sc-number)
                       (setf (tn-sc tn)
                             (svref *backend-sc-numbers* sb-vm:constant-sc-number)
                             (tn-offset tn)
@@ -803,6 +932,10 @@
 
     (do-ir2-blocks (block component)
       (emit-moves-and-coercions block))
+
+    #+arm64
+    (choose-zero-tn (ir2-component-constant-tns 2comp))
+
     ;; Give the optimizers a second opportunity to alter newly inserted vops
     ;; by looking for patterns that have a shorter expression as a single vop.
     (run-vop-optimizers component)
@@ -813,5 +946,7 @@
                     (note-if-number-stack tn 2comp ,restricted))))
       (frob ir2-component-normal-tns nil)
       (frob ir2-component-wired-tns t)
-      (frob ir2-component-restricted-tns t)))
+      (frob ir2-component-restricted-tns t)
+      #-c-stack-is-control-stack
+      (unwire-nfp-save-tn 2comp)))
   (values))
