@@ -275,61 +275,79 @@
 ;;; A (SIGNED-BYTE 64) can be represented with either fixnum or a bignum with
 ;;; exactly one digit.
 
-(define-vop (signed-byte-64-p type-predicate)
+(define-vop (pointerp)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:temporary (:sc unsigned-reg :from (:argument 0)) temp)
+  (:conditional :z)
+  (:policy :fast-safe)
+  (:translate pointerp)
+  (:generator 3
+    (if (location= temp value) (inst sub :dword value 3) (inst lea :dword temp (ea -3 value)))
+    (inst test :byte temp #b11)))
+
+;; A fixnum or single-digit bignum satisfies signed-byte-64-p
+(define-vop (signed-byte-64-p pointerp)
   (:translate signed-byte-64-p)
-  (:generator 45
-    (multiple-value-bind (yep nope)
-        (if not-p
-            (values not-target target)
-            (values target not-target))
-      (case n-fixnum-tag-bits
-        (1 (%lea-for-lowtag-test temp value other-pointer-lowtag)
-           (inst test :byte temp fixnum-tag-mask) ; 0th bit = 1 => fixnum
-           (inst jmp :nz yep)
-           (inst test :byte temp lowtag-mask))
-        (t ;; we'll only examine 1 byte, but moving a :dword is preferable
-           ;; as it doesn't require the CPU to retain the other 7 bytes.
-           (inst mov :dword temp value)
-           (inst test :byte temp fixnum-tag-mask)
-           (inst jmp :e yep)
-           (inst and :byte temp lowtag-mask)
-           (inst cmp :byte temp other-pointer-lowtag)))
-      (inst jmp :ne nope)
-      (inst cmp :qword (ea (- other-pointer-lowtag) value)
-            (+ (ash 1 n-widetag-bits) bignum-widetag))
-      (inst jmp (if not-p :ne :e) target))
-    NOT-TARGET))
+  (:conditional :z)
+  (:args-var arg-ref)
+  (:generator 6
+    (when (types-equal-or-intersect (tn-ref-type arg-ref) (specifier-type 'fixnum))
+      (inst test :byte value fixnum-tag-mask)
+      (inst jmp :z out)) ; good
+    (let ((ea (cond ((fixnum-or-other-pointer-tn-ref-p arg-ref)
+                     (ea (- other-pointer-lowtag) value))
+                    (t
+                     (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+                     (inst test :byte temp lowtag-mask)
+                     (inst jmp :nz out)
+                     (ea temp)))))
+      (inst cmp :qword ea (bignum-header-for-length 1)))
+    OUT))
+
+;;; Sign bit and fixnum tag bit.
+(defconstant non-negative-fixnum-mask-constant
+  #x8000000000000001)
+(defconstant non-negative-fixnum-mask-constant-wired-address
+  (+ static-space-start (* 12 n-word-bytes)))
+;; the preceding constant is embedded in an array,
+;; the header of which must not overlap the static alloc regions
+#-sb-thread
+(aver (>= (- non-negative-fixnum-mask-constant-wired-address (* 2 n-word-bytes))
+          (+ (max boxed-region cons-region mixed-region) (* 3 n-word-bytes))))
 
 ;;; An (unsigned-byte 64) can be represented with either a positive
 ;;; fixnum, a bignum with exactly one positive digit, or a bignum with
 ;;; exactly two digits and the second digit all zeros.
 (define-vop (unsigned-byte-64-p type-predicate)
   (:translate unsigned-byte-64-p)
-  (:generator 45
+  (:generator 10
     (let ((not-target (gen-label))
           (single-word (gen-label))
-          (fixnum (gen-label)))
+          (fixnum-p (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))))
       (multiple-value-bind (yep nope)
           (if not-p
               (values not-target target)
               (values target not-target))
-        ;; Is it a fixnum?
-        (inst mov temp value)
-        (inst test :byte temp fixnum-tag-mask)
-        (inst jmp :e fixnum)
-
-        ;; If not, is it an other pointer?
-        (inst and :byte temp lowtag-mask)
-        (inst cmp :byte temp other-pointer-lowtag)
-        (inst jmp :ne nope)
+        (when fixnum-p
+          ;; Is it a fixnum with the sign bit clear?
+          (inst test (ea non-negative-fixnum-mask-constant-wired-address) value)
+          (inst jmp :z yep))
+        (cond ((fixnum-or-other-pointer-tn-ref-p args)
+               (when fixnum-p
+                 (inst test :byte value fixnum-tag-mask)
+                 (inst jmp :z nope)))
+              (t
+               (%lea-for-lowtag-test temp value other-pointer-lowtag)
+               (inst test :byte temp lowtag-mask)
+               (inst jmp :ne nope)))
         ;; Get the header.
         (loadw temp value 0 other-pointer-lowtag)
         ;; Is it one?
-        (inst cmp temp (+ (ash 1 n-widetag-bits) bignum-widetag))
+        (inst cmp temp (bignum-header-for-length 1))
         (inst jmp :e single-word)
         ;; If it's other than two, we can't be an (unsigned-byte 64)
         ;: Leave TEMP holding 0 in the affirmative case.
-        (inst sub temp (+ (ash 2 n-widetag-bits) bignum-widetag))
+        (inst sub temp (bignum-header-for-length 2))
         (inst jmp :ne nope)
         ;; Compare the second digit to zero (in TEMP).
         (inst cmp (object-slot-ea value (1+ bignum-digits-offset) other-pointer-lowtag)
@@ -340,9 +358,7 @@
         (emit-label single-word)
         ;; Get the single digit.
         (loadw temp value bignum-digits-offset other-pointer-lowtag)
-
         ;; positive implies (unsigned-byte 64).
-        (emit-label fixnum)
         (inst test temp temp)
         (inst jmp (if not-p :s :ns) target)
 
@@ -364,8 +380,13 @@
   (:generator 4
      (let* ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
                            hi
-                           (fixnumize hi))))
-       (inst test value (constantize (lognot fixnum-hi))))))
+                           (fixnumize hi)))
+            (mask (lognot fixnum-hi))
+            (constant
+             (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
+                 (ea non-negative-fixnum-mask-constant-wired-address)
+                 (constantize mask))))
+       (inst test value constant))))
 
 (define-vop (test-fixnum-mod-tagged-unsigned)
   (:args (value :scs (any-reg unsigned-reg signed-reg)))
@@ -409,22 +430,14 @@
   (define characterp character-widetag)
   (define unbound-marker-p unbound-marker-widetag))
 
-(define-vop (pointerp)
-  (:args (value :scs (any-reg descriptor-reg) :target temp))
-  (:temporary (:sc unsigned-reg :from (:argument 0)) temp)
-  (:conditional :z)
-  (:policy :fast-safe)
-  (:translate pointerp)
-  (:generator 3
-    (inst lea :dword temp (ea -3 value))
-    (inst test :byte temp #b11)))
-
 ;;; FUNCTIONP, LISTP, %INSTANCEP, %OTHER-POINTER-P produce a flag result
 (macrolet ((define (name lowtag)
              `(define-vop (,name pointerp)
                 (:translate ,name)
                 (:generator 2
-                  (inst lea :dword temp (ea (- ,lowtag) value))
+                  (if (location= temp value)
+                      (inst sub :dword value ,lowtag)
+                      (inst lea :dword temp (ea (- ,lowtag) value)))
                   (inst test :byte temp lowtag-mask)))))
   (define functionp fun-pointer-lowtag)
   (define listp list-pointer-lowtag)
@@ -477,6 +490,18 @@
   (:translate non-null-symbol-p)
   (:generator 3 (test-other-ptr value args symbol-widetag temp out) out))
 
+(eval-when (:compile-toplevel) (aver (= sb-impl::package-id-bits 16)))
+(define-vop (keywordp symbolp)
+  (:translate keywordp)
+  (:generator 3
+    (inst lea temp (ea (- other-pointer-lowtag) value))
+    (inst test :byte temp lowtag-mask)
+    (inst jmp :ne out)
+    (inst cmp :byte (ea temp) symbol-widetag)
+    (inst jmp :ne out)
+    (inst cmp :word (ea (+ (ash symbol-name-slot word-shift) 6) temp) sb-impl::+package-id-keyword+)
+    out))
+
 (define-vop (consp type-predicate)
   (:translate consp)
   (:generator 8
@@ -514,18 +539,21 @@
    (:args (object :scs (descriptor-reg))
           (value :scs (any-reg descriptor-reg)))
    (:vop-var vop)
+   (:temporary (:sc unsigned-reg) temp)
+   (:node-var node)
    (:generator 1
+     (unless (sb-c::set-slot-old-p node)
+       (emit-gc-store-barrier object nil temp (vop-nth-arg 1 vop) value))
      (inst mov :dword (ea (- 4 instance-pointer-lowtag) object) value)))
  (define-vop (%fun-layout %instance-layout)
    (:translate %fun-layout)
    (:variant fun-pointer-lowtag))
  (define-vop (%set-fun-layout %set-instance-layout)
    (:translate %set-fun-layout)
-   (:generator 50
-     (pseudo-atomic ()
-       (inst push object)
-       (invoke-asm-routine 'call 'touch-gc-card vop)
-       (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value))))
+   (:generator 1
+     (unless (sb-c::set-slot-old-p node)
+       (emit-gc-store-barrier object nil temp (vop-nth-arg 1 vop) value))
+     (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value)))
  (define-vop ()
   (:translate sb-c::layout-eq)
   (:policy :fast-safe)
@@ -543,23 +571,28 @@
               (make-fixup (tn-value layout) :layout)
               layout)))))
 
-;;; Return the DISP part of an EA based on MEM-OP, which is comprised
-;;; of a vop name and its codegen info.
-;;; Return NIL if the value can't be encoded in a machine instruction.
-(defun valid-memref-byte-disp (mem-op)
-  (ecase (car mem-op)
-    ((instance-index-ref-c
-      raw-instance-ref-c/word
-      raw-instance-ref-c/signed-word)
-     (destructuring-bind (index) (cdr mem-op)
+;;; Return the DISP part of an EA based on MEM-OP,
+;;; which is a memory access vop such as INSTANCE-REF.
+;;; Return NIL if the access can't be absorbed into a following instruction.
+(defun valid-memref-byte-disp (mem-op &aux (info (vop-codegen-info mem-op)))
+  (ecase (vop-name mem-op)
+    (instance-index-ref-c
+     ;; for historical reasons, this has a "-C" variant which takes an info arg
+     (destructuring-bind (index) info
        (- (ash (+ index instance-slots-offset) word-shift)
           instance-pointer-lowtag)))
+    ((%raw-instance-ref/word %raw-instance-ref/signed-word)
+     ;; raw slot vops accept an immediate TN, not a codegen arg
+     (let ((index (tn-ref-tn (tn-ref-across (vop-args mem-op)))))
+       (when (sc-is index immediate)
+         (- (ash (+ (tn-value index) instance-slots-offset) word-shift)
+            instance-pointer-lowtag))))
     (slot
-     (destructuring-bind (name index lowtag) (cdr mem-op)
+     (destructuring-bind (name index lowtag) info
        (declare (ignore name))
        (- (ash index word-shift) lowtag)))
     (data-vector-ref-with-offset/simple-vector-c
-     (destructuring-bind (index offset) (cdr mem-op)
+     (destructuring-bind (index offset) info
        (let ((disp (- (ash (+ vector-data-offset index offset) word-shift)
                       other-pointer-lowtag)))
          (if (typep disp '(signed-byte 32)) disp))))))
@@ -597,7 +630,7 @@
   ;; to (memref+fixnump (memref x)), it would seem to allow matching of the pattern
   ;; again if this optimizer is reapplied, because the "new" fixnump vop is superficially
   ;; the same, except for the attachment of extra data to its input.
-  (unless (tn-ref-memory-access (sb-c::vop-args vop))
+  (unless (tn-ref-memory-access (vop-args vop))
     (let ((prev (sb-c::previous-vop-is
                  vop
                  '(instance-index-ref-c slot
@@ -613,15 +646,14 @@
                  (eq (vop-name prev) 'data-vector-ref-with-offset/simple-vector-c)
                  #-ubsan (sb-c::policy (sb-c::vop-node vop) (= safety 3)))
         (return-from vop-optimize-fixnump-optimizer nil))
-      (aver (not (sb-c::vop-results vop))) ; is a :CONDITIONAL vop
+      (aver (not (vop-results vop))) ; is a :CONDITIONAL vop
       (when (and prev (eq (vop-block prev) (vop-block vop)))
-        (let ((arg (sb-c::vop-args vop)))
-          (when (and (eq (tn-ref-tn (sb-c::vop-results prev)) (tn-ref-tn arg))
+        (let ((arg (vop-args vop)))
+          (when (and (eq (tn-ref-tn (vop-results prev)) (tn-ref-tn arg))
                      (sb-c::very-temporary-p (tn-ref-tn arg)))
-            (binding* ((mem-op (cons (vop-name prev) (vop-codegen-info prev)))
-                       (disp (valid-memref-byte-disp mem-op) :exit-if-null)
+            (binding* ((disp (valid-memref-byte-disp prev) :exit-if-null)
                        (arg-ref
-                        (sb-c::reference-tn (tn-ref-tn (sb-c::vop-args prev)) nil))
+                        (sb-c:reference-tn (tn-ref-tn (vop-args prev)) nil))
                        (new (sb-c::emit-and-insert-vop
                              (sb-c::vop-node vop) (vop-block vop) (sb-c::vop-info vop)
                              arg-ref nil prev (vop-codegen-info vop))))
@@ -629,3 +661,85 @@
               (sb-c::delete-vop prev)
               (sb-c::delete-vop vop)
               new)))))))
+
+(define-vop (>-integer-fixnum)
+  (:translate >)
+  (:args (integer :scs (descriptor-reg))
+         (fixnum :scs (immediate any-reg)))
+  (:arg-types (:or integer bignum) tagged-num)
+  (:temporary (:sc unsigned-reg) temp)
+  (:conditional)
+  (:info target not-p)
+  (:args-var args)
+  (:policy :fast-safe)
+  (:variant-vars comparison)
+  (:variant :g)
+  (:generator 8
+    (unless (sc-is (tn-ref-tn args) descriptor-reg control-stack)
+      (setf args (tn-ref-across args)))
+    (let* ((integer-p (csubtypep (tn-ref-type args) (specifier-type 'integer)))
+           (fixnum (if (sc-is fixnum immediate)
+                       (let* ((value (fixnumize (tn-value fixnum))))
+                         (cond ((plausible-signed-imm32-operand-p value)
+                                value)
+                               (t
+                                (inst mov temp value)
+                                temp)))
+                       fixnum)))
+      (multiple-value-bind (yep nope)
+          (if not-p
+              (values not-target target)
+              (values target not-target))
+        (assemble ()
+          (when (types-equal-or-intersect (tn-ref-type args) (specifier-type 'fixnum))
+            (generate-fixnum-test integer)
+            (inst jmp :nz BIGNUM)
+            (if (eql fixnum 0)
+                (inst test integer integer)
+                (inst cmp integer fixnum))
+            (inst jmp comparison yep)
+            (inst jmp nope))
+          bignum
+          (unless (fixnum-or-other-pointer-tn-ref-p args)
+            (test-type integer temp nope t (other-pointer-lowtag)))
+          (loadw temp integer 0 other-pointer-lowtag)
+          (unless integer-p
+            (inst cmp :byte temp bignum-widetag)
+            (inst jmp :ne nope))
+          #.(assert (= (integer-length bignum-widetag) 5))
+          (inst shr temp 5)
+          (inst cmp :qword (ea (- other-pointer-lowtag) integer temp) 0)
+          (inst jmp (case comparison
+                      ((:l :le) (if not-p :ge :l))
+                      (t (if not-p :l :ge)))
+                target))))
+    not-target))
+
+(define-vop (<-integer-fixnum >-integer-fixnum)
+  (:translate <)
+  (:variant :l))
+
+(define-vop (>-fixnum-integer >-integer-fixnum)
+  (:translate >)
+  (:args (fixnum :scs (immediate any-reg))
+         (integer :scs (descriptor-reg)))
+  (:arg-types tagged-num (:or integer bignum))
+  (:variant :l))
+
+(define-vop (<-fixnum-integer >-fixnum-integer)
+  (:translate <)
+  (:variant :g))
+
+;;; For integerp+cmp
+(define-vop (<=-integer-fixnum >-integer-fixnum)
+  (:translate)
+  (:variant :le))
+(define-vop (>=-integer-fixnum <-integer-fixnum)
+  (:translate)
+  (:variant :ge))
+(define-vop (<=-fixnum-integer >-fixnum-integer)
+  (:translate)
+  (:variant :le))
+(define-vop (>=-fixnum-integer <-fixnum-integer)
+  (:translate)
+  (:variant :ge))
